/* Shared library add-on to iptables to add sprio matching support. 
 * 
 * Copyright (C) 2000 Emmanuel Roger  <winfield@freegates.be>
 *
 * ChangeLog
 *     27.01.2001: Gianni Tedesco <gianni@ecsc.co.uk>
 *             Changed --tos to --sprio in save(). Also
 *             updated to work with slightly modified
 *             ipt_sprio_info.
 */

 /*
 SERCOMM   Supirior Prio  2007.05.21
-m  sprio --prio [1~n]

 */
#include <stdio.h>
#include <netdb.h>
#include <string.h>
#include <stdlib.h>
#include <getopt.h>
#include <ctype.h>

#include <iptables.h>
#include "../../../apps/fw_modules_priv/ipt_sprio.h"

/* Function which prints out usage message. */
static void
help(void)
{
	printf(
"SPRIO match v%s options:\n"
"-m sprio --prio [1~n]             Match skb->priority in a packet\n",
IPTABLES_VERSION);

	fputc('\n', stdout);
}

/*
{ "prio", 1, 0, '1' }
1st param:
  flag name
  
2nd param:
  0: no value follow the --param
  1: must have a value
  2: optional

4th param:
  it's more like an ID for --param;   this value will be used in  switch(c)   and   case'1'
  we use a char '1', which is in fact stored in corresponding int value.  
  this int will be returned as in  parse(int c,...)
*/
static struct option opts[] = { 
	{ "prio", 1, 0, '1' },
	{0}
};

/* Initialize the match. */
static void
init(struct ipt_entry_match *m, unsigned int *nfcache)
{
	*nfcache |= NFC_UNKNOWN;
}

static void
parse_sprio(const unsigned char *p, struct ipt_sprio_info *info)
{	
	u_int32_t prio; //the maximum is 65536?

	if (string_to_number(p, 0, 65535, &prio) != -1)
		info->sprio = prio;
	else
		exit_error(PARAMETER_PROBLEM, "invalid priority setting `%s' specified", p);
}

/* Function which parses command options; returns true if it
   ate an option */
static int
parse(int c, char **argv, int invert, unsigned int *flags,
      const struct ipt_entry *entry,
      unsigned int *nfcache,
      struct ipt_entry_match **match)
{
	struct ipt_sprio_info *sprioinfo = (struct ipt_sprio_info *)(*match)->data;

	switch (c) {
	case '1':
		check_inverse(optarg, &invert, &optind, 0);
		parse_sprio(argv[optind-1], sprioinfo);
		if (invert)
			sprioinfo->invert = 1;
		*flags = 1;
		break;

	default:
		return 0;
	}
	return 1;
}

static void
print_sprio(u_int32_t prio, int invert)
{

	if (invert)
		fputc('!', stdout);
	printf("%d ",prio);
}

/* Final check; must have specified --sprio. */
static void
final_check(unsigned int flags)
{
	if (!flags)
		exit_error(PARAMETER_PROBLEM,
			   "STRING match: You must specify `--prio'");
}

/* Prints out the matchinfo. */
static void
print(const struct ipt_ip *ip,
      const struct ipt_entry_match *match,
      int numeric)
{
	printf("Supirior PRIO match ");
	print_sprio(((struct ipt_sprio_info *)match->data)->sprio,	  ((struct ipt_sprio_info *)match->data)->invert);
}

/* Saves the union ipt_matchinfo in parsable form to stdout. */
static void
save(const struct ipt_ip *ip, const struct ipt_entry_match *match)
{
	printf("--sprio ");
	print_sprio(((struct ipt_sprio_info *)match->data)->sprio,
		  ((struct ipt_sprio_info *)match->data)->invert);
}

static
struct iptables_match sprio
= { NULL,
    "sprio",
    IPTABLES_VERSION,
    IPT_ALIGN(sizeof(struct ipt_sprio_info)),
    IPT_ALIGN(sizeof(struct ipt_sprio_info)),
    &help,
    &init,
    &parse,
    &final_check,
    &print,
    &save,
    opts
};

void _init(void)
{
	register_match(&sprio);
}

