/******************************************************************************
**
** FILE NAME    : amazon_se_sdio.c
** PROJECT      : Amazon SE
** MODULES      : SDIO
**
** DATE         : 06 DEC 2006
** AUTHOR       : Reddy Mallikarjuna 
** DESCRIPTION  : SDIO Driver
** COPYRIGHT    :       Copyright (c) 2006
**                      Infineon Technologies AG
**                      Am Campeon 1-12, 85579 Neubiberg, Germany
**
**    This program is free software; you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation; either version 2 of the License, or
**    (at your option) any later version.
**
** HISTORY
** $Date      		$Author     		$Comment
   06 DEC 2006          Reddy Mallikarjuna      First version
*******************************************************************************/
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/completion.h>
#include <linux/errno.h>
#include <linux/mm.h>
#include <linux/slab.h>
#include <linux/delay.h>
#include <asm/uaccess.h>
#include <linux/proc_fs.h>
#include <net/sock.h>
#include <linux/skbuff.h>
#include <linux/netlink.h>

#include <asm/amazon_se/amazon_se.h>
#include <asm/amazon_se/amazon_se_dma.h>
#include <asm/amazon_se/irq.h>
#include <asm/io.h>

#include <asm/amazon_se/infineon_sdio_card.h>
#include <asm/amazon_se/infineon_sdio.h>
#include <asm/amazon_se/infineon_sdio_cmds.h>
#include <asm/amazon_se/infineon_sdio_controller.h>
#include <asm/amazon_se/amazon_se_sdio_controller.h>
#include <asm/amazon_se/amazon_se_sdio_controller_registers.h>

#include <linux/mmc/mmc.h>
#include <linux/delay.h>


#define AMAZON_SE_SDIO_CONTROLLER_INTR0_IRQ  INT_NUM_IM1_IRL20
#define AMAZON_SE_SDIO_CONTROLLER_INTR1_IRQ  INT_NUM_IM1_IRL21
#define AMAZON_SE_SDIO_CONTROLLER_I_IRQ INT_NUM_IM1_IRL22

#define AMAZON_SE_SDIO_VERSION "1.0.0"


#define GPIO_MODE_SET		1

// MCLCMD       GPIO 19
#define MCLCMD		19
// MCLCLK       GPIO 15 
#define MCLCLK		15
// MCLDATA0     GPIO 12 
#define MCLDATA0	12
// MCLDATA1     GPIO 26 
#define MCLDATA1	26
// MCLDATA2     GPIO 28 
#define MCLDATA2	28
// MCLDATA3     GPIO 20 
#define MCLDATA3	20

#define DMA_CLASS 0x0

//#define AMAZON_SE_SDIO_DEBUG
#undef	AMAZON_SE_SDIO_DEBUG
//#undef AMAZON_SE_SDIO_DMSG
//#define AMAZON_SE_SDIO_DMSG
#ifdef AMAZON_SE_SDIO_DEBUG
#define AMAZON_SE_SDIO_DMSG(fmt, args...) printk( KERN_INFO  "%s: " fmt,__FUNCTION__, ## args)
#else
#define AMAZON_SE_SDIO_DMSG(fmt, args...)
#endif

#define AMAZON_SE_SDIO_EMSG(fmt, args...) printk( KERN_ERR  "%s: " fmt,__FUNCTION__, ## args)

MODULE_DESCRIPTION ("AMAZON_SE sdio controller module");
MODULE_AUTHOR ("Reddy Mallikarjuna");

#define SD_BLOCK_SIZE 512
#define DMA_RESET

#ifdef CONFIG_INFINEON_SDIO
static int amazon_se_sd_controller_send_data (struct sd_controller *pDev,
					   sdio_card_t * pCard,
					   sd_block_request_t * pRequest,
					   uint32_t timeout);
static int amazon_se_sd_controller_read_data_pre (struct sd_controller *pDev,
					       sdio_card_t * pCard,
					       sd_block_request_t * pRequest);
static int amazon_se_sd_controller_read_data (struct sd_controller *pDev,
					   sdio_card_t * pCard,
					   sd_block_request_t * pRequest,
					   uint32_t timeout);
static int amazon_se_sd_controller_send_cmd (struct sd_controller *pDev,
					  struct sd_cmd *pCmd);
#endif
static void amazon_se_sd_controller_sdio_mask_and_ack_irq (void);
static int amazon_se_sd_controller_set_ops (int type, uint32_t data);
static int setup_dma_driver (amazon_se_sd_controller_priv_t * priv);
static int cleanup_dma_driver (amazon_se_sd_controller_priv_t * priv);

#ifdef CONFIG_INFINEON_SDIO
static struct completion comp_amazon_se_sd_controller_cmd;
static wait_queue_head_t wait_queue_amazon_se_sd_controller_cmd;
static wait_queue_head_t wait_queue_amazon_se_sd_controller_data_cmd;
static struct completion comp_amazon_se_sd_controller_data_cmd;
static sd_block_request_t *data_request = NULL;
static int  data_access = 0;
#endif 
static int cmd_access = 0;

#ifdef CONFIG_TI_SDIO
static int      g_data_recv_len=0;
static int      g_data_len=0;
static int      g_data_len_offset=0;
static void (*g_sdio_completion_callback)(unsigned long);
uint8_t *g_sdio_async_receive_buffer = NULL;
uint8_t *g_rx_receive_buffer = NULL;
static int  tx_data_err= 0;
#endif //CONFIG_TI_SDIO

#define AMAZON_SE_SDIO_MAJOR 201
static int major = AMAZON_SE_SDIO_MAJOR;
#ifdef CONFIG_INFINEON_SDIO
static uint32_t decode_taac (uint8_t taac);
#endif //CONFIG_INFINEON_SDIO

static sdio_controller_t amazon_se_sd_controller = {
	.name = DEVICE_NAME,
#ifdef CONFIG_INFINEON_SDIO
	.send_data = amazon_se_sd_controller_send_data,
	.read_data_pre = amazon_se_sd_controller_read_data_pre,
	.read_data = amazon_se_sd_controller_read_data,
	.send_cmd = amazon_se_sd_controller_send_cmd,
	.mask_ack_irq = amazon_se_sd_controller_sdio_mask_and_ack_irq,
#endif
	.set_ops = amazon_se_sd_controller_set_ops,
	//.VDD          = 0xffff00,
	.VDD = (BIT20 | BIT21),	//onlys support 3.3V
	.blklen = 9,
};

static int amazon_se_sdio_ioctl (struct inode *ino, struct file *fil,
			      unsigned int command, unsigned long lon);

static struct file_operations amazon_se_sdio_operations = {
      ioctl:amazon_se_sdio_ioctl,
};

#define AMAZON_SE_SDIO_PROC_DIRNAME     "amazon_se_sdio"
#define PROC_ITEMS 12

static struct proc_dir_entry *amazon_se_sdio_dir;

static int amazon_se_sdio_proc_read (struct file *file, char *buf, size_t nbytes,
				  loff_t * ppos);

static struct file_operations proc_operations = {
      read:amazon_se_sdio_proc_read,
};

typedef struct reg_entry {
	int *flag;
	char name[30];		// big enough to hold names
	char description[100];	// big enough to hold description
	unsigned short low_ino;
} reg_entry_t;
uint32_t cmd_response_counter = 0;
uint32_t cmd_counter = 0;
uint32_t cmd_error_counter = 0;
uint32_t data_read_counter = 0;
uint32_t data_read_total_size = 0;
uint32_t data_write_counter = 0;
uint32_t data_write_total_size = 0;
uint32_t data_error_counter = 0;


//uint32_t data_read_total_size = 0;
uint32_t data_read_async_counter = 0;
uint32_t data_read_async_total_size = 0;
//uint32_t data_write_counter = 0;
//uint32_t data_write_total_size = 0;
uint32_t data_write_async_counter = 0;
uint32_t data_write_async_total_size = 0;
//uint32_t data_error_counter = 0;
uint32_t data_tx_error_counter = 0;
uint32_t data_rx_error_counter = 0;
uint32_t cmd_crc_error_counter=0;
uint32_t cmd_crc_rx_error_counter=0;
uint32_t cmd_crc_tx_error_counter=0;
uint32_t cmd_tx_timeout_counter=0;
uint32_t cmd_rx_timeout_counter=0;

static reg_entry_t regs[PROC_ITEMS];	// total items to be monitored by /proc/mei

#define NUM_OF_REG_ENTRY	(sizeof(regs)/sizeof(reg_entry_t))

#define AMAZON_SE_SD_CONTROLLER_WRITEL(data,addr)      do{ *((volatile u32*)(addr)) = (u32)(data); asm("SYNC");} while (0)
#define AMAZON_SE_SD_CONTROLLER_READL(addr)    (*((volatile u32*)(addr)))

#ifdef CONFIG_INFINEON_SDIO
/**
    Send several blocks data to sd card
\param pDev	ptr to actual SD Controller
\param pCard	ptr to card device
\param pRequest	ptr to the request command
\return 	IFX_SUCCESS or IFX_ERROR if send command fail
\remark		Send a block data to sd card
*/
static int
amazon_se_sd_controller_send_data (struct sd_controller *pDev,
				sdio_card_t * pCard,
				sd_block_request_t * pRequest,
				uint32_t timeout)
{
	amazon_se_sd_controller_priv_t *priv = pDev->priv;
	struct dma_device_info *dma_dev = priv->dma_device;
	uint32_t reg = 0;
	uint32_t data_ctrl_reg;
	uint8_t *pData = NULL;
	block_data_t *blk_data;
	int need_to_alloc = 0;
	uint32_t clk_timeout;

#if 0
	if (pDev->card_transfer_state != pCard) {
		return ERROR_WRONG_CARD_STATE;
	}
#endif
	data_write_counter++;
	data_request = pRequest;
	dma_dev->current_tx_chan = 0;
	data_access = 1;
	AMAZON_SE_SD_CONTROLLER_WRITEL (pRequest->request_size, MCI_DLGTH);
	data_ctrl_reg = (pRequest->block_size_shift << 4) | 0x9;	//blksize,dir=ctrl to card,dma enable 
	AMAZON_SE_SD_CONTROLLER_WRITEL (AMAZON_SE_SD_CONTROLLER_READL (MCI_IM0) |
				     MCI_IM_DTO | MCI_IM_DBE | MCI_IM_DCF,
				     MCI_IM0);
	if (pCard) {
		clk_timeout =
			pCard->csd.nsac * 100 * 100 +
			(uint32_t) ((decode_taac (pCard->csd.taac) / 10) *
				    (amazon_se_sd_controller.current_speed /
				     10000) / 100);
		clk_timeout = clk_timeout << pCard->csd.r2w_factor;
	}
	else {
		clk_timeout = 0xFFFFFFFF;
	}
	AMAZON_SE_SD_CONTROLLER_WRITEL (clk_timeout, MCI_DTIM);
	AMAZON_SE_SD_CONTROLLER_WRITEL (data_ctrl_reg, MCI_DCTRL);
	blk_data = pRequest->blocks;
	pData = pRequest->blocks->data;

	//make sure the address is continuous
	while (pRequest->blocks) {
		if (pRequest->blocks->next) {
			if (pRequest->blocks->data +
			    (1 << pRequest->block_size_shift) !=
			    pRequest->blocks->next->data) {
				need_to_alloc = 1;
			}
		}
		pRequest->blocks = pRequest->blocks->next;
	}
	if (need_to_alloc) {
		int idx = 0;
		char *pData_copy;
		pData_copy = kmalloc (pRequest->request_size, GFP_ATOMIC);
		if (pData_copy == NULL) {
			data_error_counter++;
			return -ENOMEM;
		}
		while (blk_data) {
			memcpy (pData_copy + idx, blk_data->data,
				(1 << pRequest->block_size_shift));
			idx += (1 << pRequest->block_size_shift);
			blk_data = blk_data->next;
		}
		kfree (pData);
		pData = pData_copy;
	}
	if (dma_device_write (dma_dev, pData, pRequest->request_size, NULL) !=
	    pRequest->request_size) {
		printk ("buffer full\n");
		pRequest->error = -EBUSY;
		data_error_counter++;
		goto send_data_end;
	}

	data_write_total_size += pRequest->request_size;
	if (data_access)
		interruptible_sleep_on_timeout
			(&wait_queue_amazon_se_sd_controller_data_cmd, timeout);
	if (data_access != 0) {
		AMAZON_SE_SDIO_EMSG ("data access timeout\n");
		pRequest->error = ERROR_TIMEOUT;
		data_access = 0;
	}
	if (pRequest->error) {
		printk ("Error:pRequest->error=%d\n", pRequest->error);
		goto send_data_end;
	}
      send_data_end:
	data_request = NULL;
	reg = AMAZON_SE_SD_CONTROLLER_READL (MCI_IM0);
	reg &= ~(MCI_STAT_DTO | MCI_STAT_DBE);
	AMAZON_SE_SD_CONTROLLER_WRITEL (reg, MCI_IM0);
	return pRequest->error;
}
#endif //CONFIG_INFINEON_SDIO

#ifdef CONFIG_TI_SDIO
static uint32_t uiTxCmdMciStat1;
static uint32_t uiRxCmdMciStat1;
static int rStatSync=0, wStatSync=0,rStatAsync=0,wStatAsync=0;
//int amazon_se_sdio_set_flag(int active);
int amazon_se_sd_controller_send_data_kernel(struct sd_cmd *pCmd, void *buffer, int length)
{
        int i, ret;
        uint32_t *pData = buffer ;
        int  cnt;
        unsigned long flags;
        uint32_t counter;
        local_irq_save(flags);
//      AMAZON_SE_SD_CONTROLLER_WRITEL (((0x6<<4)| 0x4), MCI_DCTRL);
        ret = amazon_se_sd_controller_send_cmd_kernel(pCmd);
        if (pCmd->error == MMC_ERR_TIMEOUT)
        {
                local_irq_restore(flags);
                AMAZON_SE_SDIO_EMSG("Tx Command Time out(Sync)\n");
                cmd_tx_timeout_counter++;
                return ret;
        }
        AMAZON_SE_SD_CONTROLLER_WRITEL (0x7ff, MCI_CL);
        AMAZON_SE_SD_CONTROLLER_WRITEL (0xfffff, MCI_DTIM);
        AMAZON_SE_SD_CONTROLLER_WRITEL (length, MCI_DLGTH);
        AMAZON_SE_SD_CONTROLLER_WRITEL (((0x6<<4)| 0x5), MCI_DCTRL);
        for (i=0;(i<length/4) && (i < 16 );i++)
        {
                 *((volatile uint32_t *)MCI_DF0)  = *pData++;
        }
	for (;i<length/4;i++)
        {
                counter=0xfffff;
                while(!((*(volatile uint32_t *)MCI_STAT) & MCI_STAT_TXHF))
                {
                        counter--;
                        if(counter == 0)
                        break;
                }

                if(counter == 0) break;
                 *((volatile uint32_t*)MCI_DF0) = *pData++;
        }
        counter=0xfffff;
        while (((*(volatile uint32_t *)MCI_STAT) & MCI_STAT_TXA)) {   //fix me
                counter--;
                if(counter == 0) break;
        }  //fix me

/*         amazon_se_sdio_set_flag(1);  */
        /* 8 clock cycles of delay is required before the return */
        udelay(1);
//      ndelay(350);  /* @25MHz, 1 cycle = 40ns, min 40*(32/4)=320ns*/
//      ndelay(650);  /* @12.5MHz, 1 cycle = 80ns, min 80*(32/4)=640ns*/

         local_irq_restore(flags);
cnt = *(volatile uint32_t *)MCI_DCNT;
        if(cnt != 0)
        {
                data_error_counter++;
                data_tx_error_counter++;
                AMAZON_SE_SDIO_EMSG("%d: MCI_DCNT:%d , Status:%0x8 (Data time out, Tx Sync),Length:%d \n",__LINE__,cnt,(*(volatile uint32_t *)MCI_STAT), length);
                AMAZON_SE_SD_CONTROLLER_WRITEL (((0x6<<4)| 0x4), MCI_DCTRL);
//              ret |= MMC_ERR_TIMEOUT;
/*
        counter=0xfffff;
                while (!(*(volatile uint32_t *)MCI_STAT & MCI_STAT_DTO)); //fix me
                {
                        counter--;
                        if(counter == 0) break;
                }
*/
        }
        else
        {
                data_write_total_size += length;

        }
        * (volatile uint32_t *)MCI_CL = 0x7ff; /* clear all status bits */
        data_write_counter++;
//        local_irq_restore(flags);
        if (pCmd->error == MMC_ERR_BADCRC)
        {
                cmd_crc_tx_error_counter++;
        }
        wStatSync=ret;  //Debug  info
//      AMAZON_SE_SD_CONTROLLER_WRITEL (((0x6<<4)|0x4), MCI_DCTRL);

        return ret;
}

int dma_device_write_pre (struct dma_device_info *dma_dev)
{
        unsigned long  flags  ;
        _dma_channel_info *pCh;
         int chan_no;
        struct tx_desc *tx_desc_p;
        u32 cs_val;
        pCh = dma_dev->tx_chan[dma_dev->current_tx_chan];
        tx_desc_p = (struct tx_desc *) pCh->desc_base + pCh->curr_desc;
//      if (tx_desc_p->status.field.OWN == 1 /*DMA_OWN*/)
        {
//              chan_no = (int) (pCh - (_dma_channel_info *) dma_chan);

//                dma_dev->buffer_free ((u8 *) __va (tx_desc_p->Data_Pointer),NULL);
                cs_val = *AMAZON_SE_DMA_CS;
                local_irq_save(flags);
                *AMAZON_SE_DMA_CS = 15;
                *AMAZON_SE_DMA_CCTRL &= ~1;        //Channel off

                *AMAZON_SE_DMA_CCTRL |= 2;
                while (*AMAZON_SE_DMA_CCTRL & 2);
                 *AMAZON_SE_DMA_IRNEN |= 1 << 15;
                 *AMAZON_SE_DMA_CCTRL = 0x10100;
//               *AMAZON_SE_DMA_CCTRL = 0x00100;
//               *AMAZON_SE_DMA_CIE = 0x0A;
                pCh->desc_len = 1;
                *AMAZON_SE_DMA_CDBA = (u32) CPHYSADDR (pCh->desc_base);
               *AMAZON_SE_DMA_CDLEN = pCh->desc_len;
                *AMAZON_SE_DMA_CS = cs_val;
                local_irq_restore(flags);
                memset (tx_desc_p, 0, sizeof (struct tx_desc));
                tx_data_err=0;
        }

}


int amazon_se_sd_controller_send_async_data_kernel(struct sd_cmd *pCmd, void *buffer, int length)
{
        int ret;
        struct sd_controller * dev = &amazon_se_sd_controller;
        int  cnt;
        unsigned long flags;
        uint32_t counter;
        uint32_t reg = 0;
        unsigned long status = 0;
        amazon_se_sd_controller_priv_t *priv = dev->priv;
        struct dma_device_info *dma_dev = priv->dma_device;
        dma_dev->current_tx_chan = 0;

//      AMAZON_SE_SD_CONTROLLER_WRITEL (((0x6<<4)| 0xc), MCI_DCTRL);
        AMAZON_SE_SD_CONTROLLER_WRITEL ((SDIO_DMACON_TXON | DMA_CLASS), SDIO_DMACON);
        local_irq_save(flags);
        ret = amazon_se_sd_controller_send_cmd_kernel(pCmd);
        if (pCmd->error == MMC_ERR_TIMEOUT)
        {
                local_irq_restore(flags);
                AMAZON_SE_SDIO_EMSG("Tx Command Time out (async operation)\n");
                cmd_tx_timeout_counter++;
                status = (unsigned long )ret;
                if (g_sdio_completion_callback != 0)
                            g_sdio_completion_callback(status);
                return ret;
        }
        AMAZON_SE_SD_CONTROLLER_WRITEL (0x7ff, MCI_CL);
        AMAZON_SE_SD_CONTROLLER_WRITEL (0xfffff, MCI_DTIM);
        AMAZON_SE_SD_CONTROLLER_WRITEL (length, MCI_DLGTH);
        local_irq_restore(flags);
        AMAZON_SE_SD_CONTROLLER_WRITEL (((0x6<<4)|0x0d), MCI_DCTRL);
 local_irq_restore(flags);
        AMAZON_SE_SD_CONTROLLER_WRITEL (((0x6<<4)|0x0d), MCI_DCTRL);
        if (dma_device_write (dma_dev, (uint8_t*)buffer, length, NULL) != length)
        {
                AMAZON_SE_SDIO_EMSG("ERROR: buffer full(async operation)!!!\n");
                data_error_counter++;
                data_tx_error_counter++;
                dma_device_write_pre(dma_dev);
//              ret|= MMC_ERR_TIMEOUT;  /*Error condition */
        }


//      local_irq_save(flags);
        counter=0xfffff;
        while (((*(volatile uint32_t *)MCI_STAT) & MCI_STAT_TXA)) {   //fix me
//      while( *(volatile uint32_t *)MCI_DCNT !=0) {
                counter--;
                if(counter == 0) break;
        }  //fix me

        /* 8 clock cycles of delay is required before the return */
        udelay(1);
//      ndelay(350);  /* @25MHz, 1 cycle = 40ns, min 40*(32/4)=320ns*/
//      ndelay(650);  /* @12.5MHz, 1 cycle = 80ns, min 80*(32/4)=640ns*/

//      AMAZON_SE_SD_CONTROLLER_WRITEL (0x0, SDIO_DMACON);
//      local_irq_restore(flags);
        cnt = *(volatile uint32_t *)MCI_DCNT;
 if(cnt != 0)
        {
        //      AMAZON_SE_SDIO_EMSG("%d: MCI_DCNT:%d (Data time out, Tx ASync), Length:%d, counter:%08x \n",__LINE__,cnt,length,counter);
        //      printk("response[0] = %08x staus:%08x, guard_time=%d us\n", pCmd->response[0], (*(volatile uint32_t *)MCI_STAT), guard_time);
                AMAZON_SE_SD_CONTROLLER_WRITEL (((0x6<<4)| 0xc), MCI_DCTRL);
                if(counter == 0)
                printk(" response[0] = %08x Status:%08x\n", pCmd->response[0], (*(volatile uint32_t *)MCI_STAT) );
                data_error_counter++;
                data_tx_error_counter++;
                dma_device_write_pre(dma_dev);
//              ret |= MMC_ERR_TIMEOUT;
        }
        else
        {
                data_write_total_size += length;
                data_write_async_total_size += length;
        }

        data_write_counter++;
        data_write_async_counter++;
        data_write_total_size += length;
        data_write_async_total_size += length;

        if (pCmd->error == MMC_ERR_BADCRC)
        {
                cmd_crc_tx_error_counter++;
        }

        wStatAsync=ret;  //Debug info
//      AMAZON_SE_SD_CONTROLLER_WRITEL (((0x6<<4)|0xc), MCI_DCTRL);
        status = (unsigned long )ret;
        AMAZON_SE_SD_CONTROLLER_WRITEL (0x7ff, MCI_CL);
        if (g_sdio_completion_callback != 0)
            g_sdio_completion_callback(status);
  return ret;
}
#endif //CONFIG_TI_SDIO


#ifdef CONFIG_INFINEON_SDIO
/**
    Read several blocks data from sd card
\param pDev	ptr to actual SD Controller
\param pCard	ptr to card device
\param pRequest	ptr to the request command
\return - IFX_SUCCESS 
	- IFX_ERROR if an error occured
\remark	Read a block data to sd card
 */
static int
amazon_se_sd_controller_read_data_pre (struct sd_controller *pDev,
				    sdio_card_t * pCard,
				    sd_block_request_t * pRequest)
{
	uint32_t data_ctrl_reg;

#if 0
	if (pDev->card_transfer_state != pCard) {
		return ERROR_WRONG_CARD_STATE;
	}
#endif

	AMAZON_SE_SDIO_DMSG
		("pRequest->nBlocks=%d,pRequest->request_size=%d,block_size_shift=%d\n",
		 pRequest->nBlocks, pRequest->request_size,
		 pRequest->block_size_shift);
	data_request = pRequest;

// dma
	AMAZON_SE_SD_CONTROLLER_WRITEL (0xFFFFffff, MCI_DTIM);	// need to check
	AMAZON_SE_SD_CONTROLLER_WRITEL (AMAZON_SE_SD_CONTROLLER_READL (MCI_IM0) |
				     MCI_IM_DTO | MCI_STAT_DCF | MCI_IM_SBE,
				     MCI_IM0);
	data_ctrl_reg = (pRequest->block_size_shift << 4) | 0xB;	//blksize,dir=card to ctrl,dma enable 

	AMAZON_SE_SD_CONTROLLER_WRITEL ((1 << pRequest->block_size_shift) *
				     pRequest->nBlocks, MCI_DLGTH);
	AMAZON_SE_SDIO_DMSG ("DLENGTH=%d\n",
			  AMAZON_SE_SD_CONTROLLER_READL (MCI_DLGTH));

	data_access = 1;
	AMAZON_SE_SD_CONTROLLER_WRITEL (data_ctrl_reg, MCI_DCTRL);

	return OK;
}

static uint32_t
decode_taac (uint8_t taac)
{
	uint32_t ret = 0;
	static const unsigned int tacc_exp[] = {
		1, 10, 100, 1000, 10000, 100000, 1000000, 10000000,
	};

	static const unsigned int tacc_mant[] = {
		0, 10, 12, 13, 15, 20, 25, 30,
		35, 40, 45, 50, 55, 60, 70, 80,
	};

	ret = (tacc_exp[taac & 0x7] * tacc_mant[(taac >> 3) & 0xF] + 9) / 10;
	return ret;
}

static int
amazon_se_sd_controller_read_data (struct sd_controller *pDev,
				sdio_card_t * pCard,
				sd_block_request_t * pRequest,
				uint32_t timeout)
{
	uint32_t reg = 0;
	uint32_t clk_timeout = 0;
	uint32_t data_size = pRequest->request_size;

	AMAZON_SE_SDIO_DMSG ("Data reading,total length=%d!\n",
			  pRequest->request_size);
#if 0
	if (pDev->card_transfer_state != pCard) {
		return ERROR_WRONG_CARD_STATE;
	}
#endif
	if (pCard) {
		clk_timeout =
			pCard->csd.nsac * 100 * 100 +
			(uint32_t) ((decode_taac (pCard->csd.taac) / 10) *
				    (amazon_se_sd_controller.current_speed /
				     10000) / 100);
	}
	else {
		clk_timeout = 0xFFFFFFFF;
	}

	AMAZON_SE_SD_CONTROLLER_WRITEL (clk_timeout, MCI_DTIM);
	data_read_counter++;

	if (data_access == 1) {
		// wait for reading
		interruptible_sleep_on_timeout
			(&wait_queue_amazon_se_sd_controller_data_cmd, timeout);

	}
	// read finish
	reg = AMAZON_SE_SD_CONTROLLER_READL (MCI_IM0);
	reg &= ~(MCI_IM_DTO | MCI_STAT_DCF);
	AMAZON_SE_SD_CONTROLLER_WRITEL (reg, MCI_IM0);
	if (data_access != 0) {
		data_request = NULL;
		AMAZON_SE_SDIO_EMSG ("Data read timeout!\n");
		data_access = 0;
		data_error_counter++;
		return ERROR_TIMEOUT;
	}
	else {
		data_read_total_size += data_size;
	}
	data_request = NULL;
	return pRequest->error;
}
#endif //CONFIG_INFINEON_SDIO

#ifdef CONFIG_TI_SDIO
int dma_device_read_pre_reset (struct dma_device_info *dma_dev, int len)
{
        int byte_offset = 0;
        unsigned long  flags  ;
        _dma_channel_info *pCh = dma_dev->rx_chan[dma_dev->current_rx_chan];
        struct rx_desc *rx_desc_p;
        u32 cs_val = *AMAZON_SE_DMA_CS;
//      u32 ps_val = *AMAZON_SE_DMA_PS;
        local_irq_save(flags);
//      *AMAZON_SE_DMA_PS = 3;
  //      *AMAZON_SE_DMA_PCTRL = 0x28;
        *AMAZON_SE_DMA_CS = 0xe;
        *AMAZON_SE_DMA_CCTRL &= ~1;        //Channel off

        *AMAZON_SE_DMA_CCTRL |= 2;
        while (*AMAZON_SE_DMA_CCTRL & 2);
         *AMAZON_SE_DMA_IRNEN |= 1 << 0xe;
         *AMAZON_SE_DMA_CCTRL = 0x10100;
       //  *AMAZON_SE_DMA_CCTRL = 0x00100;
         *AMAZON_SE_DMA_CIE = 0x0A;
        pCh->desc_len = 1;

       *AMAZON_SE_DMA_CDLEN = pCh->desc_len;
//      printk("current desc:%d\n", pCh->curr_desc);
        rx_desc_p = (struct rx_desc *) pCh->desc_base + pCh->curr_desc;
        rx_desc_p->status.word = 0;
        if(len )
        {
        rx_desc_p->status.field.byte_offset =  byte_offset;
        rx_desc_p->status.field.data_length = len;
        rx_desc_p->status.field.OWN = 1;        //DMA_OWN;
        *AMAZON_SE_DMA_CS = 0xe;
        *AMAZON_SE_DMA_CCTRL |= 1; //Channel on
        }
        *AMAZON_SE_DMA_CS = cs_val;
//      *AMAZON_SE_DMA_PS = ps_val;
        local_irq_restore(flags);
       return 1;
}

int dma_device_read_pre_len (struct dma_device_info *dma_dev, int len)
{
        int byte_offset = 0;
        unsigned long  flags  ;
        _dma_channel_info *pCh = dma_dev->rx_chan[dma_dev->current_rx_chan];
        struct rx_desc *rx_desc_p;
        u8* buffer;
        u32 cs_val;
#if 1
         buffer = dma_dev->buffer_alloc (len, &byte_offset, NULL);
        if(buffer == NULL)
        {
                printk("Buffer is NULL\n");
        }
#endif
        rx_desc_p = (struct rx_desc *) pCh->desc_base + pCh->curr_desc;
        rx_desc_p->Data_Pointer = (u32) CPHYSADDR ((u32) buffer);
//        rx_desc_p->Data_Pointer = (u32) CPHYSADDR ((u32)g_rx_receive_buffer);
        rx_desc_p->status.word = 0;
        rx_desc_p->status.field.byte_offset =  byte_offset;
        rx_desc_p->status.field.data_length = len;
        rx_desc_p->status.field.OWN = 1;        //DMA_OWN;

#ifndef CONFIG_MIPS_UNCACHED
          dma_cache_inv ((unsigned long) buffer,len);
//          dma_cache_inv ((unsigned long) g_rx_receive_buffer,len);
#endif
//      printk("Byte offset:%d, len:%d, status word: %08x, Data Pointer:%08X, buffer:%08x\n",byte_offset,len,rx_desc_p->status.word,rx_desc_p->Data_Pointer,buffer);
        return 1;
}

int dump_data_and_time(void *buffer, int length)
{
    int i;
    uint32_t *pData = buffer;
    printk("Tx Command response Tx: 0x%08x/ Rx: 0x%08x \n", uiTxCmdMciStat1,uiRxCmdMciStat1 );
    printk("Cmd crc error counter (rx+tx+cmd): 0x%08x\n", cmd_crc_error_counter);
    printk("Cmd crc Rx error counter: 0x%08x\n", cmd_crc_rx_error_counter);
    printk("Cmd crc Tx error counter: 0x%08x\n", cmd_crc_tx_error_counter);
    printk("Data read counter: 0x%08x\n",data_read_counter );
    printk("Data read counter: 0x%08x\n",data_read_async_counter );
    printk("Data read total size: 0x%08x\n", data_read_total_size);
    printk("Data read total size: 0x%08x\n", data_read_async_total_size);
    printk("Data write counter: 0x%08x\n", data_write_counter);
    printk("Data write counter: 0x%08x\n", data_write_async_counter);
    printk("Data write total size: 0x%08x\n", data_write_total_size);
    printk("Data write total size: 0x%08x\n", data_write_async_total_size);
    printk("Data Tx error counter: 0x%08x\n", data_tx_error_counter);
    printk("Data Rx error counter: 0x%08x\n", data_rx_error_counter);
    printk("Data error counter (rx+tx): 0x%08x\n", data_error_counter);
    return 0;
}
int amazon_se_sd_controller_read_data_kernel(struct sd_cmd *pCmd, void *buffer, int length)
{
        int i,rTry=0;
        int ret=0;
        uint32_t *pData = buffer;
        unsigned long  flags  ;
        int cnt, temp;
        uint32_t  counter,  stat  ;
#if 0
        memset (buffer, 0, length);
#endif
        AMAZON_SE_SD_CONTROLLER_WRITEL (0x7ff, MCI_CL);
        AMAZON_SE_SD_CONTROLLER_WRITEL (AMAZON_SE_SD_CONTROLLER_READL(SDIO_DMACON) & ~SDIO_DMACON_RXON, SDIO_DMACON);
        local_irq_save(flags);
        AMAZON_SE_SD_CONTROLLER_WRITEL (0xfffff, MCI_DTIM);
        AMAZON_SE_SD_CONTROLLER_WRITEL (length, MCI_DLGTH);
        AMAZON_SE_SD_CONTROLLER_WRITEL (((0x6<<4)|0x7), MCI_DCTRL);
        ret = amazon_se_sd_controller_send_cmd_kernel(pCmd);
        if (pCmd->error == MMC_ERR_TIMEOUT)
        {
                local_irq_restore(flags);
                AMAZON_SE_SDIO_EMSG("Rx Command Time out (Sync)\n");
                cmd_rx_timeout_counter++;
                return ret;
        }
        i=0;
        counter=0xfffff;
        temp = length;
 while((temp!=0)&& counter--  )
        {

                        cnt = *(volatile uint32_t *)MCI_DCNT;
                while ((((temp-cnt) >= 8) || ((temp == 4) && (cnt == 0))) && (!(*(volatile uint32_t *)MCI_STAT & MCI_STAT_RXFE)) )

                {
                                *pData++ = * (volatile uint32_t *)((uint32_t)MCI_DF0);
                                temp -=4;

                }
        }


        local_irq_restore(flags);
        data_read_counter++;
        counter=0xffff;
        cnt = *(volatile uint32_t *)MCI_DCNT;
        if(cnt!=0 )
        {
                data_error_counter++;
                data_rx_error_counter++;
                AMAZON_SE_SDIO_EMSG("%d: MCI_DCNT:%d , Status:%0x8 (Data time out, Tx Sync),Length:%d \n",__LINE__,cnt,(*(volatile uint32_t *)MCI_STAT), length);
                AMAZON_SE_SD_CONTROLLER_WRITEL (((0x6<<4)|0x6), MCI_DCTRL);
//              while (!(*(volatile uint32_t *)MCI_STAT & MCI_STAT_DTO)); //Fix me
//              ret |= MMC_ERR_TIMEOUT;
/*
                {
                        counter--;
                        if(counter == 0) break;
                }

*/
 }
        else
        {
                data_read_total_size += length;

        }
        if (pCmd->error == MMC_ERR_BADCRC)
        {
                        cmd_crc_rx_error_counter++;
        }


//      AMAZON_SE_SD_CONTROLLER_WRITEL (((0x6<<4)|0x6), MCI_DCTRL);

        AMAZON_SE_SD_CONTROLLER_WRITEL (0x7ff, MCI_CL);
        rStatSync=ret;  //Debug info
        return ret;

}
static int reset_dma=1;
int amazon_se_sd_controller_read_async_data_kernel(struct sd_cmd *pCmd, void *buffer, int length)
{
        int ret=0;
        unsigned long  flags  ;
        unsigned long status = 0;
        uint32_t reg=0, tmplen=length;
        uint32_t  counter;
        int cnt ;

//      static unsigned int min_guard_time_rx = 300;
//      AMAZON_SE_SD_CONTROLLER_WRITEL (((0x6<<4)|0xe), MCI_DCTRL);
        AMAZON_SE_SD_CONTROLLER_WRITEL (AMAZON_SE_SD_CONTROLLER_READL(SDIO_DMACON)|(SDIO_DMACON_RXON| DMA_CLASS), SDIO_DMACON);
        g_sdio_async_receive_buffer = buffer;

        struct sd_controller * dev = &amazon_se_sd_controller;
      amazon_se_sd_controller_priv_t *priv = dev->priv;
      struct dma_device_info *dma_dev = priv->dma_device;
        g_data_len_offset= 0;
        g_data_recv_len=0;
#ifndef DMA_RESET
        if(reset_dma)
        {
                dma_device_read_pre_reset (dma_dev, 0);
                reset_dma=0;
        }
#endif
        if (length %16 == 0)
        {
#ifdef DMA_RESET
                dma_device_read_pre_reset (dma_dev, tmplen);
#else


                dma_device_read_pre_len (dma_dev, tmplen);
#endif
}
        else
        {
                tmplen = ((length/16)*16)+16;
                g_data_len_offset = 16-(length%16);
#ifdef DMA_RESET
                dma_device_read_pre_reset (dma_dev, tmplen);
#else
                dma_device_read_pre_len (dma_dev, tmplen);
#endif
        }
//        g_data_len= length;
        g_data_len=  tmplen;
        AMAZON_SE_SD_CONTROLLER_WRITEL (0x7ff, MCI_CL);

        AMAZON_SE_SD_CONTROLLER_WRITEL (( MCI_IM_DTO | MCI_IM_DBE | MCI_IM_DCF|MCI_IM_SBE|MCI_IM_RO),    MCI_IM1);


        AMAZON_SE_SD_CONTROLLER_WRITEL (0xfffff, MCI_DTIM);
//      AMAZON_SE_SD_CONTROLLER_WRITEL (length, MCI_DLGTH);
        AMAZON_SE_SD_CONTROLLER_WRITEL (tmplen, MCI_DLGTH);
        local_irq_save(flags);
        AMAZON_SE_SD_CONTROLLER_WRITEL (((0x6<<4)|0xF), MCI_DCTRL);
        ret = amazon_se_sd_controller_send_cmd_kernel(pCmd);

        local_irq_restore(flags);
        if (pCmd->error == MMC_ERR_TIMEOUT)
        {
                AMAZON_SE_SDIO_EMSG("Rx Command Time out(async operation)\n");
                cmd_rx_timeout_counter++;
                status = (unsigned long )ret;
//              return OK;
                if (g_sdio_completion_callback != 0)
                            g_sdio_completion_callback(status);
                return ret;
        }
data_read_counter++;
        data_read_async_counter++;



#if 0
//      local_irq_save(flags);
        counter=0xfffff;
        while (((*(volatile uint32_t *)MCI_STAT) & MCI_STAT_RXA)) {   //fix me
//      while( *(volatile uint32_t *)MCI_DCNT !=0) {
                counter--;
                if(counter == 0) break;
        }  //fix me

        cnt = *(volatile uint32_t *)MCI_DCNT;
        if(cnt != 0)
        {
                if(counter == 0)
                {
                printk("Counter is zero");
                printk(" response[0] = %08x Status:%08x\n", pCmd->response[0], (*(volatile uint32_t *)MCI_STAT) );
                }
                else
                printk("Counter is not zero");
                data_rx_error_counter++;
                dma_device_read_pre_reset (dma_dev, 0);
//              ret |= MMC_ERR_TIMEOUT;
        }
        else
        {
//              printk(" response[0] = %08x Status:%08x\n", pCmd->response[0], (*(volatile uint32_t *)MCI_STAT) );
//              AMAZON_SE_SDIO_EMSG("%d: MCI_DCNT:%d  Length:%d, counter:%08x \n",__LINE__,cnt,length,counter);
        }
#endif
if (pCmd->error == MMC_ERR_BADCRC)
        {
                        cmd_crc_rx_error_counter++;
        }



        rStatAsync=ret;  //Debug info
        return ret;
}


/**
    Send a command to sd card
\param pDev	ptr to actual SD Controller
\param pCmd	ptr to the request command
\return - IFX_SUCCESS 
	- IFX_ERROR if an error occured
\remark	Send a command to sd card
 */
int amazon_se_sd_controller_send_cmd_kernel (struct sd_cmd *pCmd)
{
    int ret;
    ret = amazon_se_sd_controller_send_cmd_sync(&amazon_se_sd_controller, pCmd);
     if ((ret == 1)|| (ret == -1 ))
     printk("%s:%s:%d Leave, ret=%d\n", __FILE__, __FUNCTION__, __LINE__, ret);
    return ret;
}
static int amazon_se_sd_controller_handle_cmd_int (struct sd_controller *pDev);
static int
amazon_se_sd_controller_send_cmd_sync (struct sd_controller *pDev,
                               struct sd_cmd *pCmd)
{
        uint32_t sd_cmd = 0;
        uint32_t reg = 0, counter;
        int nRetry = 0;
        int ret;

      send_retry:

        AMAZON_SE_SDIO_DMSG ("Sending command :op_code=%d,type=0x%X args=0x%X\n",
                          pCmd->op_code, pCmd->response_type, pCmd->args);

        AMAZON_SE_SD_CONTROLLER_WRITEL (pCmd->args, MCI_ARG);
        sd_cmd = pCmd->op_code;
        switch (pCmd->response_type) {
        case SD_RSP_R1:
        case SD_RSP_R1b:
        case SD_RSP_R3:
        case SD_RSP_R4:
        case SD_RSP_R5:
        case SD_RSP_R6:
                sd_cmd |= MCI_CMD_SHORT_RSP;
                AMAZON_SE_SD_CONTROLLER_WRITEL (AMAZON_SE_SD_CONTROLLER_READL
                                             (MCI_IM0) | MCI_IM_CRE |
                                             MCI_IM_CTO | MCI_IM_CCF,
                                             MCI_IM0);
                break;
        case SD_RSP_R2:
                sd_cmd |= MCI_CMD_LONG_RSP;
                AMAZON_SE_SD_CONTROLLER_WRITEL (AMAZON_SE_SD_CONTROLLER_READL
                                             (MCI_IM0) | MCI_IM_CRE |
                                             MCI_IM_CTO | MCI_IM_CCF,
                                             MCI_IM0);
                break;
 default:
                AMAZON_SE_SD_CONTROLLER_WRITEL (AMAZON_SE_SD_CONTROLLER_READL
                                             (MCI_IM0) | MCI_IM_CS, MCI_IM0);
                break;
        }
        pDev->cmd = pCmd;

        pDev->cmd->error = OK;
        sd_cmd |= MCI_CMD_EN;

        cmd_access = 1;

        AMAZON_SE_SD_CONTROLLER_WRITEL (sd_cmd, MCI_CMD);
        cmd_counter++;

        counter = 500;
        {
            do {
                ret = amazon_se_sd_controller_handle_cmd_int(pDev);
            } while (((ret == -1) && counter--) );
        }
        if(counter == 0) printk("c ");

        reg = AMAZON_SE_SD_CONTROLLER_READL (MCI_IM0);
        reg &= ~(MCI_IM_CRE | MCI_IM_CTO | MCI_IM_CCF | MCI_IM_CS);
        AMAZON_SE_SD_CONTROLLER_WRITEL (reg, MCI_IM0);

        if (cmd_access != 0) {
                AMAZON_SE_SDIO_DMSG ("cmd access timeout\n");
                pDev->cmd->error = ERROR_TIMEOUT;
                cmd_access = 0;
                nRetry++;
                if (nRetry < 2)
                        goto send_retry;
        }

        return ret;
}

/* return: 0 - no wakeup ; 1 - wakeup ; -1 - error*/
static int
amazon_se_sd_controller_handle_cmd_int (struct sd_controller *pDev)
{
        u32 status, irq_done = 0;
        struct sd_cmd *cmd = pDev->cmd;
        u32 reg_im0 = 0;
        int ret = -1;

        status = AMAZON_SE_SD_CONTROLLER_READL (MCI_STAT);

        reg_im0 = AMAZON_SE_SD_CONTROLLER_READL (MCI_IM0);
        status &= reg_im0;

        //Command Sent
        //Command was sent. No response required.
        if (status & MCI_STAT_CS) {
                if (cmd->response_type == SD_RSP_NONE) {
                        cmd_access = 0;
                        ret = OK; /* wakeup */
                }
        }

        //Command Response End
        //Command response received. CRC check passed.
        if (status & MCI_STAT_CRE) {
                if (AMAZON_SE_SD_CONTROLLER_READL (MCI_REPCMD) ==
                    pDev->cmd->op_code) {

                        switch (cmd->response_type) {
                        case SD_RSP_R1:
                        case SD_RSP_R1b:
                        case SD_RSP_R3:
                        case SD_RSP_R4:
                        case SD_RSP_R5:
case SD_RSP_R6:
                                cmd->response[0] =
                                        AMAZON_SE_SD_CONTROLLER_READL (MCI_REP0);
                                cmd_response_counter++;
                                AMAZON_SE_SDIO_DMSG ("got interrupt,resp=%08X\n",
                                                  cmd->response[0]);
                                break;
                        default:
                                cmd->error = ERROR_WRONG_RESPONSE_TYPE;
                                break;
                        }
                        cmd_access = 0;
                        ret = OK; /* wakeup */

                }
                else {
                        if (AMAZON_SE_SD_CONTROLLER_READL (MCI_REPCMD) == 0x3F && cmd->response_type == SD_RSP_R2) // Response CMD of Response R2 is 3F
                        {

                                cmd->response[0] =
                                        AMAZON_SE_SD_CONTROLLER_READL (MCI_REP0);
                                cmd->response[1] =
                                        AMAZON_SE_SD_CONTROLLER_READL (MCI_REP1);
                                cmd->response[2] =
                                        AMAZON_SE_SD_CONTROLLER_READL (MCI_REP2);
                                cmd->response[3] =
                                        AMAZON_SE_SD_CONTROLLER_READL (MCI_REP3);
                                cmd_access = 0;
                                cmd_response_counter++;
                                AMAZON_SE_SDIO_DMSG
                                        ("got interrupt,resp[0]=0x%8X,resp[1]=0x%08X,resp[2]=0x%08X,resp[3]=0x%08X\n",
                                         cmd->response[0], cmd->response[1],
                                         cmd->response[2], cmd->response[3]);
                                ret = OK; /* wakeup */
                        }
 else {
                                //break;
                                cmd->error = ERROR_WRONG_RESPONSE_CMD;
                        }
                }
        }
        //Command Time-out
        //Command time-out indication.
        if (status & MCI_STAT_CTO)      // command time-out | command crc fail
        {
//            amazon_se_sdio_set_flag(0);
 //           amazon_se_sdio_set_flag(1);
//              cmd->error = ERROR_TIMEOUT;
                cmd->error = MMC_ERR_TIMEOUT;   //2
//              ret = -1;
                ret = MMC_ERR_TIMEOUT;  //Command Time out, TI error code  /* linux/mmc/mmc.h*/
        }

        //Command CRC Fail
        //Command response received, CRC check failed.
        if (status & MCI_STAT_CCF) {
                if (cmd->response_type == SD_RSP_R3 && AMAZON_SE_SD_CONTROLLER_READL (MCI_REPCMD) == 0x3F) // Response R3 CRC is always 7F, Response CMD is always 3F
                {
                        cmd->response[0] =
                                AMAZON_SE_SD_CONTROLLER_READL (MCI_REP0);
                        AMAZON_SE_SDIO_DMSG ("got interrupt,resp=%08X\n",
                                          cmd->response[0]);
                       ret = OK;        /* wakeup */
                }
                else {
//                      cmd->error = ERROR_CRC_ERROR;
                        cmd->error = MMC_ERR_BADCRC;  //1
                        cmd_crc_error_counter++;
 ret = MMC_ERR_BADCRC;  //Command Crc error,  TI error code  /* linux/mmc/mmc.h*/
                }
                cmd_access = 0;
//               ret = OK;        /* wakeup */
        }

        AMAZON_SE_SD_CONTROLLER_WRITEL (status, MCI_CL);
        return ret;
}

static void
amazon_se_sd_controller_handle_intr1 (struct sd_controller *pDev)
{
        u32 status ;
        struct sd_cmd *cmd = pDev->cmd;
        u32 reg_im1 = 0;         u32 stat = 0;         uint32_t reg = 0;
        int ret =0;
   /*  ASYNC CALLBACK */
        unsigned long ti_status = 0; /* error status, as used by TI (see amazon_se_sdiodrv.c report_error) */
        struct sd_controller * dev = &amazon_se_sd_controller;
        amazon_se_sd_controller_priv_t *priv = dev->priv;
        struct dma_device_info *dma_dev = priv->dma_device;
        dma_dev->current_tx_chan = 0;


        status = AMAZON_SE_SD_CONTROLLER_READL (MCI_STAT);
        stat=status;
        AMAZON_SE_SDIO_DMSG ("Got interrupt, STATUS=%08X\n", status);
        reg_im1 = AMAZON_SE_SD_CONTROLLER_READL (MCI_IM1);
        status &= reg_im1;


        //Data Block End
        //Data block sent. CRC check passed.
        if (status & MCI_STAT_DBE) {
                AMAZON_SE_SDIO_EMSG("MCI_STAT_DBE, never comes to here in stream mode\n");
                if (AMAZON_SE_SD_CONTROLLER_READL (MCI_DCNT) == 0) {
                        ti_status |= OK;
                }
                ret = 1;
        }
        if (status & MCI_STAT_RO) {
                u32 cnt1, cnt2;
 ti_status = OK;
                ret = 1;
                cnt1 = *(volatile uint32_t *)MCI_DCNT;
                u32 counter=0xfffff;
                while (((*(volatile uint32_t *)MCI_STAT) & MCI_STAT_RXA)) {   //fix me
                counter--;
                if(counter == 0)
                         break;
                }  //fix me
//              cnt2 = *(volatile uint32_t *)MCI_DCNT;
//              AMAZON_SE_SDIO_EMSG("MCI_STAT_RO,Status:%08x , counter:%08x, cnt1: %d, cnt2: %d\n",stat,counter,cnt1,cnt2);
                AMAZON_SE_SD_CONTROLLER_WRITEL (((0x6<<4)|0xe), MCI_DCTRL);
                ndelay(240);
//              stat = AMAZON_SE_SD_CONTROLLER_READL (MCI_STAT);
//              AMAZON_SE_SDIO_EMSG("MCI_STAT_RO,Status:%08x , counter:%08x, cnt1: %d, cnt2: %d\n",stat,counter,cnt1,cnt2);
                AMAZON_SE_SD_CONTROLLER_WRITEL (AMAZON_SE_SD_CONTROLLER_READL(SDIO_DMACON) & ~SDIO_DMACON_RXON, SDIO_DMACON);
                dma_device_read_pre_reset (dma_dev, 0);
                data_rx_error_counter++;
        }


        //Start Bit Error
        //Start bit not detected on all data signals in wide bus mode.
        if (status & MCI_STAT_SBE) {
                AMAZON_SE_SDIO_EMSG ("Error:MCI_STAT_SBE\n");
//              data_request->error = ERROR_CRC_ERROR;
                ti_status |= MMC_ERR_BADCRC;
        }

        if (status & MCI_STAT_TU) {

                u32 cnt;
//              ti_status |= MMC_ERR_TIMEOUT;
ti_status = OK;
                tx_data_err = 1;
                dma_device_write_pre(dma_dev);
                cnt = *(volatile uint32_t *)MCI_DCNT;
                AMAZON_SE_SDIO_EMSG("MCI_STAT_TU MCI_DCNT:%d\n",cnt);
                ret = 1;
//              if (g_sdio_completion_callback != 0)
//                  g_sdio_completion_callback(ti_status);

        }
        //Data End
        //Data end indication means that the data counter is zero.
        if (status & MCI_STAT_DE) {
        //      uint32_t counter;
//                      AMAZON_SE_SDIO_EMSG("MCI_STAT_DE,wake up!\n");
#if 1
//              udelay(1);
                ret = 1;
                ti_status = OK;
//        if (g_sdio_completion_callback != 0)
  //          g_sdio_completion_callback(ti_status);
#endif
/*
                counter=0xfffff;
                while (((*(volatile uint32_t *)MCI_STAT) & MCI_STAT_TXA)) {   //fix me
                        counter--;
                        if(counter == 0) break;
                }  //fix me
*/
        }


 //Data Time-out
        //Data time-out indication.
        if (status & MCI_STAT_DTO)      // data timeout
        {
                AMAZON_SE_SDIO_EMSG ("Data Timeout.\n");
                ti_status |= MMC_ERR_TIMEOUT;
                AMAZON_SE_SD_CONTROLLER_WRITEL (AMAZON_SE_SD_CONTROLLER_READL(SDIO_DMACON) & ~SDIO_DMACON_RXON, SDIO_DMACON);
                ret =1;
                ti_status = OK;
//              if (g_sdio_completion_callback != 0)
  //                  g_sdio_completion_callback(ti_status);
        }


        //Data CRC Fail
        //Data block send/received CRC check failed.
        if (status & MCI_STAT_DCF) {
                AMAZON_SE_SDIO_EMSG ("ERROR_CRC_ERROR! MCI_STAT_DCF\n");
                ti_status |= MMC_ERR_BADCRC;
                ret =1;
        }
        if(ret)
        {
        AMAZON_SE_SD_CONTROLLER_WRITEL (status, MCI_CL);
//      AMAZON_SE_SD_CONTROLLER_WRITEL (0x7ff, MCI_CL);
        reg = AMAZON_SE_SD_CONTROLLER_READL (MCI_IM1);
        reg &= ~(MCI_IM_DTO | MCI_IM_DBE|MCI_IM_DCF|MCI_IM_DE |MCI_IM_SBE|MCI_IM_TU | MCI_IM_RO);
        AMAZON_SE_SD_CONTROLLER_WRITEL (reg, MCI_IM1);
//      AMAZON_SE_SD_CONTROLLER_WRITEL (AMAZON_SE_SD_CONTROLLER_READL(SDIO_DMACON) & ~SDIO_DMACON_RXON, SDIO_DMACON);
   /*  ASYNC CALLBACK */
#if  1
        if (g_sdio_completion_callback != 0)
                    g_sdio_completion_callback(ti_status);
#endif
 }

//      return OK;
}
#endif //CONFIG_TI_SDIO

#ifdef CONFIG_INFINEON_SDIO
/**
    Send a command to sd card
\param pDev	ptr to actual SD Controller
\param pCmd	ptr to the request command
\return - IFX_SUCCESS 
	- IFX_ERROR if an error occured
\remark	Send a command to sd card
 */
static int
amazon_se_sd_controller_send_cmd (struct sd_controller *pDev,
			       struct sd_cmd *pCmd)
{
	uint32_t sd_cmd = 0;
	uint32_t reg = 0;
	int nRetry = 0;

      send_retry:

	AMAZON_SE_SDIO_DMSG ("Sending command :op_code=%d,type=0x%X args=0x%X\n",
			  pCmd->op_code, pCmd->response_type, pCmd->args);

	AMAZON_SE_SD_CONTROLLER_WRITEL (pCmd->args, MCI_ARG);
	sd_cmd = pCmd->op_code;
	switch (pCmd->response_type) {
	case SD_RSP_R1:
	case SD_RSP_R1b:
	case SD_RSP_R3:
	case SD_RSP_R4:
	case SD_RSP_R5:
	case SD_RSP_R6:
		sd_cmd |= MCI_CMD_SHORT_RSP;
		AMAZON_SE_SD_CONTROLLER_WRITEL (AMAZON_SE_SD_CONTROLLER_READL
					     (MCI_IM0) | MCI_IM_CRE |
					     MCI_IM_CTO | MCI_IM_CCF,
					     MCI_IM0);
		break;
	case SD_RSP_R2:
		sd_cmd |= MCI_CMD_LONG_RSP;
		AMAZON_SE_SD_CONTROLLER_WRITEL (AMAZON_SE_SD_CONTROLLER_READL
					     (MCI_IM0) | MCI_IM_CRE |
					     MCI_IM_CTO | MCI_IM_CCF,
					     MCI_IM0);
		break;
	default:
		AMAZON_SE_SD_CONTROLLER_WRITEL (AMAZON_SE_SD_CONTROLLER_READL
					     (MCI_IM0) | MCI_IM_CS, MCI_IM0);
		break;
	}
	pDev->cmd = pCmd;

	pDev->cmd->error = OK;
	sd_cmd |= MCI_CMD_EN;

	cmd_access = 1;

	AMAZON_SE_SD_CONTROLLER_WRITEL (sd_cmd, MCI_CMD);
	cmd_counter++;

	interruptible_sleep_on_timeout (&wait_queue_amazon_se_sd_controller_cmd,
					100);

	reg = AMAZON_SE_SD_CONTROLLER_READL (MCI_IM0);
	reg &= ~(MCI_IM_CRE | MCI_IM_CTO | MCI_IM_CCF | MCI_IM_CS);
	AMAZON_SE_SD_CONTROLLER_WRITEL (reg, MCI_IM0);

	if (cmd_access != 0) {
		AMAZON_SE_SDIO_DMSG ("cmd access timeout\n");
		pDev->cmd->error = ERROR_TIMEOUT;
		cmd_access = 0;
		nRetry++;
		if (nRetry < 2)
			goto send_retry;
	}

	return pDev->cmd->error;
}

/**
    SD Controller interrupt handle
\param pDev	ptr to actual SD Controller
\return - IFX_SUCCESS 
	- IFX_ERROR if an error occured
\remark	SD Controller interrupt handle
 */
static int
amazon_se_sd_controller_handle_int (struct sd_controller *pDev)
{
	u32 status, irq_done = 0;
	struct sd_cmd *cmd = pDev->cmd;
	u32 reg_im0 = 0;

	status = AMAZON_SE_SD_CONTROLLER_READL (MCI_STAT);
	AMAZON_SE_SDIO_DMSG ("Got interrupt, STATUS=%08X\n", status);
	reg_im0 = AMAZON_SE_SD_CONTROLLER_READL (MCI_IM0);
	irq_done = status & (~reg_im0);
	status &= reg_im0;

	irq_done = 0;

	//Receive Data Available
	//Data available in receive FIFO.
	if (status & MCI_STAT_RXDA) {
		irq_done |= MCI_STAT_RXDA;
	}

	//Transmit Data Available
	//Data available in transmit FIFO.
	if (status & MCI_STAT_TXDA) {
		irq_done |= MCI_STAT_TXDA;
	}

	//Receive FIFO Empty
	//Returns the Receive FIFO status.
	if (status & MCI_STAT_RXFE) {
		irq_done |= MCI_STAT_RXFE;
		printk ("MCI_STAT_RXFE\n");
	}

	//Transmit FIFO Empty
	//Returns the Receive FIFO status.
	if (status & MCI_STAT_TXFE) {
		irq_done |= MCI_STAT_TXFE;
	}

	//Receive FIFO Full
	//Returns the Receive FIFO status.
	if (status & MCI_STAT_RXFF) {
		irq_done |= MCI_STAT_RXFF;
		printk ("MCI_STAT_RXFF\n");
	}

	//Transmit FIFO Full
	//Returns the Transmit FIFO status.
	if (status & MCI_STAT_TXFF) {
		irq_done |= MCI_STAT_TXFF;
	}

	//Receive FIFO half full
	//Returns the FIFO status.
	if (status & MCI_STAT_RXHF) {
		irq_done |= MCI_STAT_RXHF;
		printk ("MCI_STAT_RXHF\n");
	}

	//Transmit FIFO half empty
	//Returns the FIFO status.
	if (status & MCI_STAT_TXHF) {
		irq_done |= MCI_STAT_TXHF;
	}

	//Receive Active
	//Data receive in progress.
	if (status & MCI_STAT_RXA) {
		printk ("MCI_STAT_RXA\n");
		irq_done |= MCI_STAT_RXA;
	}

	//Transmit Active
	//Data transmit in progress.
	if (status & MCI_STAT_TXA) {
		irq_done |= MCI_STAT_TXA;
	}

	//Command Active
	//Command transfer in progress.
	if (status & MCI_STAT_CMDA) {
		irq_done |= MCI_STAT_CMDA;
	}

	//Data Block End
	//Data block sent. CRC check passed.
	if (status & MCI_STAT_DBE) {
		irq_done |= MCI_STAT_DBE;
		//AMAZON_SE_SDIO_EMSG("MCI_STAT_DBE\n");
		if (AMAZON_SE_SD_CONTROLLER_READL (MCI_DCNT) == 0) {
			data_access = 0;
			//AMAZON_SE_SDIO_EMSG("MCI_STAT_DBE,wake up!\n");
			wake_up_interruptible
				(&wait_queue_amazon_se_sd_controller_data_cmd);
		}
	}

	//Start Bit Error               
	//Start bit not detected on all data signals in wide bus mode.
	if (status & MCI_STAT_SBE) {
		irq_done |= MCI_STAT_SBE;
		AMAZON_SE_SDIO_EMSG ("Error:MCI_STAT_SBE\n");
		data_request->error = ERROR_CRC_ERROR;
		wake_up_interruptible
			(&wait_queue_amazon_se_sd_controller_data_cmd);
	}

	//Data End
	//Data end indication means that the data counter is zero.
	if (status & MCI_STAT_DE) {
		irq_done |= MCI_STAT_DE;
	}

	//Command Sent
	//Command was sent. No response required.
	if (status & MCI_STAT_CS) {
		if (cmd->response_type == SD_RSP_NONE) {
			cmd_access = 0;
			wake_up_interruptible
				(&wait_queue_amazon_se_sd_controller_cmd);
		}
		irq_done |= MCI_STAT_CS;
	}

	//Command Response End
	//Command response received. CRC check passed.
	if (status & MCI_STAT_CRE) {
		if (AMAZON_SE_SD_CONTROLLER_READL (MCI_REPCMD) ==
		    pDev->cmd->op_code) {

			switch (cmd->response_type) {
			case SD_RSP_R1:
			case SD_RSP_R1b:
			case SD_RSP_R3:
			case SD_RSP_R4:
			case SD_RSP_R5:
			case SD_RSP_R6:
				cmd->response[0] =
					AMAZON_SE_SD_CONTROLLER_READL (MCI_REP0);
				cmd_response_counter++;
				AMAZON_SE_SDIO_DMSG ("got interrupt,resp=%08X\n",
						  cmd->response[0]);
				break;
			default:
				cmd->error = ERROR_WRONG_RESPONSE_TYPE;
				break;
			}
			cmd_access = 0;
			wake_up_interruptible
				(&wait_queue_amazon_se_sd_controller_cmd);

		}
		else {
			if (AMAZON_SE_SD_CONTROLLER_READL (MCI_REPCMD) == 0x3F && cmd->response_type == SD_RSP_R2)	// Response CMD of Response R2 is 3F
			{

				cmd->response[0] =
					AMAZON_SE_SD_CONTROLLER_READL (MCI_REP0);
				cmd->response[1] =
					AMAZON_SE_SD_CONTROLLER_READL (MCI_REP1);
				cmd->response[2] =
					AMAZON_SE_SD_CONTROLLER_READL (MCI_REP2);
				cmd->response[3] =
					AMAZON_SE_SD_CONTROLLER_READL (MCI_REP3);
				cmd_access = 0;
				cmd_response_counter++;
				AMAZON_SE_SDIO_DMSG
					("got interrupt,resp[0]=0x%8X,resp[1]=0x%08X,resp[2]=0x%08X,resp[3]=0x%08X\n",
					 cmd->response[0], cmd->response[1],
					 cmd->response[2], cmd->response[3]);
				wake_up_interruptible
					(&wait_queue_amazon_se_sd_controller_cmd);
			}
			else {
				//break;
				cmd->error = ERROR_WRONG_RESPONSE_CMD;
			}
		}
		irq_done |= MCI_STAT_CRE;
	}

	//Receive Overrun
	//Receive FIFO overrun error.
	if (status & MCI_STAT_RO) {
		irq_done |= MCI_STAT_RO;
		data_request->error = ERROR_DATA_ERROR;
	}

	//Transmit Underrun
	//Transmit FIFO underrun error.
	if (status & MCI_STAT_TU) {
		irq_done |= MCI_STAT_TU;
		data_request->error = ERROR_DATA_ERROR;
	}

	//Data Time-out
	//Data time-out indication.
	if (status & MCI_STAT_DTO)	// data timeout
	{
		AMAZON_SE_SDIO_EMSG ("Data Timeout.\n");
		data_request->error = ERROR_TIMEOUT;
		wake_up_interruptible
			(&wait_queue_amazon_se_sd_controller_data_cmd);
		irq_done |= MCI_STAT_DTO;
	}

	//Command Time-out
	//Command time-out indication.
	if (status & MCI_STAT_CTO)	// command time-out | command crc fail
	{
		AMAZON_SE_SDIO_DMSG ("Command time-out | command crc fai.\n");
		cmd->error = ERROR_TIMEOUT;
		irq_done |= MCI_STAT_CTO;
	}

	//Data CRC Fail
	//Data block send/received CRC check failed.
	if (status & MCI_STAT_DCF) {
		irq_done |= MCI_STAT_DCF;
		AMAZON_SE_SDIO_EMSG ("ERROR_CRC_ERROR! MCI_STAT_DCF\n");
		data_request->error = ERROR_CRC_ERROR;
	}

	//Command CRC Fail
	//Command response received, CRC check failed.
	if (status & MCI_STAT_CCF) {
		if (cmd->response_type == SD_RSP_R3 && AMAZON_SE_SD_CONTROLLER_READL (MCI_REPCMD) == 0x3F)	// Response R3 CRC is always 7F, Response CMD is always 3F
		{
			cmd->response[0] =
				AMAZON_SE_SD_CONTROLLER_READL (MCI_REP0);
			AMAZON_SE_SDIO_DMSG ("got interrupt,resp=%08X\n",
					  cmd->response[0]);
		}
		else {
			AMAZON_SE_SDIO_EMSG
				("Command response received.Crc check failed!\n");
			cmd->error = ERROR_CRC_ERROR;
		}
		cmd_access = 0;
		wake_up_interruptible (&wait_queue_amazon_se_sd_controller_cmd);
		irq_done |= MCI_STAT_CCF;
	}

	//AMAZON_SE_SD_CONTROLLER_WRITEL(irq_done,MCI_CL);
	AMAZON_SE_SD_CONTROLLER_WRITEL (status, MCI_CL);
	if (cmd->error != OK || (data_request && data_request->error != OK)) {
		//AMAZON_SE_SDIO_EMSG("Error!!,cmd->error=%d,data_request=%d\n",cmd->error,data_request?data_request->error:999);
		goto ERROR;
	}
	//status &= ~(irq_done);

	//AMAZON_SE_SD_CONTROLLER_WRITEL(AMAZON_SE_SD_CONTROLLER_READL(SDIO_ICR)|SDIO_ICR_INTR0,SDIO_ICR);
	return OK;
      ERROR:
	//AMAZON_SE_SD_CONTROLLER_WRITEL(AMAZON_SE_SD_CONTROLLER_READL(SDIO_ICR)|SDIO_ICR_INTR0,SDIO_ICR);
	if (cmd->error != OK) {
		cmd_error_counter++;
		return cmd->error;
	}
	if (data_request && data_request->error != OK) {
		data_error_counter++;
		return data_request->error;
	}
	return -1;
}

/**
    SDIO interrupt handle
\param irq irq number
\param dev_id ptr to the Device 
\param pt_regs ptr to the registers
\return - IFX_SUCCESS 
	- IFX_ERROR if an error occured
\remark	SD Controller interrupt handle
 */
static void
amazon_se_sd_controller_handle_sdio_int (int irq, void *dev_id,
				      struct pt_regs *regs)
{
	uint32_t imc = 0;
	unsigned long flags;
	struct sd_controller *pDev = (struct sd_controller *) dev_id;
	u32 status = AMAZON_SE_SD_CONTROLLER_READL (SDIO_MIS);

	AMAZON_SE_SDIO_DMSG ("got sdio interrupt,irq=%d.\n", irq);
	save_flags (flags);
	cli ();
	if (status & SDIO_MIS_SDIO) {
		// disable sdio interrupt until sdio inturrupt finish
		imc = AMAZON_SE_SD_CONTROLLER_READL (SDIO_IMC);
		imc &= ~(SDIO_IMC_SDIO);
		AMAZON_SE_SD_CONTROLLER_WRITEL (imc, SDIO_IMC);
#ifdef SDIO_SUPPORT
		sd_core_sdio_int (pDev);
#endif
		AMAZON_SE_SD_CONTROLLER_WRITEL (AMAZON_SE_SD_CONTROLLER_READL
					     (SDIO_ICR) | SDIO_ICR_INTR0,
					     SDIO_ICR);
	}
	restore_flags (flags);

}

/**
    Mask and ACK SDIO interrupt
\return - IFX_SUCCESS 
	- IFX_ERROR if an error occured
\remark	    Mask and ACK SDIO interrupt
 */
static void
amazon_se_sd_controller_sdio_mask_and_ack_irq (void)
{
	AMAZON_SE_SD_CONTROLLER_WRITEL (AMAZON_SE_SD_CONTROLLER_READL (SDIO_ICR) |
				     SDIO_ICR_SDIO, SDIO_IMC);
}

/**
    SD interrupt handle
\param irq irq number
\param dev_id ptr to the Device 
\param pt_regs ptr to the registers
\return - IFX_SUCCESS 
	- IFX_ERROR if an error occured
\remark	SD Controller interrupt handle
 */
static void
amazon_se_sd_controller_irq (int irq, void *dev_id, struct pt_regs *regs)
{
	struct sd_controller *pDev = (struct sd_controller *) dev_id;
	unsigned long flags;

	save_flags (flags);
	cli ();

	amazon_se_sd_controller_handle_int (pDev);
	restore_flags (flags);
}
#endif //CONFIG_INFINEON_SDIO

#ifdef CONFIG_TI_SDIO
static void
amazon_se_sd_controller_irq_intr1 (int irq, void *dev_id, struct pt_regs *regs)
{
        struct sd_controller *pDev = (struct sd_controller *) dev_id;
        unsigned long flags;

        save_flags (flags);
        cli ();

        amazon_se_sd_controller_handle_intr1 (pDev);
        restore_flags (flags);
}
#endif


/**
   Chagne the VDD or FREQUENCY 
\param type type of operation
\param data data to update
\return - IFX_SUCCESS 
	- IFX_ERROR if an error occured
\remark	SD Controller interrupt handle
 */
static int
amazon_se_sd_controller_set_ops (int type, uint32_t data)
{
	uint32_t reg = 0;
	uint32_t div = 0;
	switch (type) {
	case SD_SET_VDD:
		break;
	case SD_SET_FREQENCY:
		reg = AMAZON_SE_SD_CONTROLLER_READL (MCI_CLK);
		reg &= ~(MCI_CLK_BY | 0xFF);	//BYPASS, CLK_DIV
		if (data == 0) {
			AMAZON_SE_SDIO_DMSG ("Warning! The clock is 0.\n");
			reg &= ~(MCI_CLK_EN);
			AMAZON_SE_SD_CONTROLLER_WRITEL (reg, MCI_CLK);
			amazon_se_sd_controller.current_speed = 0;
			break;
		}

		if (data >=
		    ((amazon_se_sd_controller_priv_t *) amazon_se_sd_controller.
		     priv)->mclk_speed) {
			if (data > ((amazon_se_sd_controller_priv_t *)
				    amazon_se_sd_controller.priv)->mclk_speed)
				AMAZON_SE_SDIO_EMSG
					("Error! Clock is too large. clock = %d, amazon_se mclk_speed=%d\n",
					 data,
					 ((amazon_se_sd_controller_priv_t *)
					  amazon_se_sd_controller.priv)->
					 mclk_speed);

			div = MCI_CLK_BY;
		}
		else {
			div = ((amazon_se_sd_controller_priv_t *)
			       amazon_se_sd_controller.priv)->mclk_speed /
				(data * 2);
			div -= 1;
			// make sure the clock <= data
			if ((((amazon_se_sd_controller_priv_t *)
			      amazon_se_sd_controller.priv)->mclk_speed / ((div +
									 1) *
									2)) >
			    data)
				div++;
		}

		AMAZON_SE_SDIO_DMSG ("div=%d,mclk_speed=%d\n", div,
				  ((amazon_se_sd_controller_priv_t
				    *) (amazon_se_sd_controller.priv))->
				  mclk_speed);

		reg |= div;
		reg |= MCI_CLK_EN;
	//	reg=0x101;//for venus
		AMAZON_SE_SD_CONTROLLER_WRITEL (reg, MCI_CLK);
		if (div == MCI_CLK_BY)
			amazon_se_sd_controller.current_speed =
				((amazon_se_sd_controller_priv_t *)
				 amazon_se_sd_controller.priv)->mclk_speed;
		else
			amazon_se_sd_controller.current_speed =
				((amazon_se_sd_controller_priv_t *)
				 amazon_se_sd_controller.priv)->mclk_speed /
				((div + 1) * 2);
		AMAZON_SE_SDIO_DMSG ("current_speed = %08X\n",
				  amazon_se_sd_controller.current_speed);

		AMAZON_SE_SDIO_DMSG ("\n1MCI_CLK=%08X\n",
				  AMAZON_SE_SD_CONTROLLER_READL (MCI_CLK));
		break;
	case SD_SET_BUS_WIDTH:
		reg = AMAZON_SE_SD_CONTROLLER_READL (MCI_CLK);
		switch (data) {
		case SD_BUS_4BITS:
			reg |= MCI_CLK_WD;
			break;
		case SD_BUS_1BITS:
			reg &= ~(MCI_CLK_WD);
			break;
		default:
			printk("\nReturn _EINVAL");
			return -EINVAL;
		}
		AMAZON_SE_SD_CONTROLLER_WRITEL (reg, MCI_CLK);
		AMAZON_SE_SDIO_DMSG ("MCI_CLK=%08X\n",
				  AMAZON_SE_SD_CONTROLLER_READL (MCI_CLK));
		break;
	case SD_SET_BLOCK_LEN:
		if (data != amazon_se_sd_controller.blklen) {
			amazon_se_sd_controller_priv_t *priv;
			priv = amazon_se_sd_controller.priv;
			cleanup_dma_driver (priv);
			amazon_se_sd_controller.blklen = data;
			setup_dma_driver (priv);
		}
		break;
	}

	return OK;
}

static int
amazon_se_data_recv (struct dma_device_info *dma_dev)
{
	uint8_t *buf = NULL;
	int len = 0, i = 0;
	unsigned long status = 0;

#ifdef DMA_RESET
	len = dma_device_read (dma_dev, &buf, NULL);
//      len = dma_device_read (dma_dev, &g_rx_receive_buffer, NULL);
#else
	len = dma_device_read_sdio (dma_dev, &buf, NULL);
//      len = dma_device_read_sdio (dma_dev, &g_rx_receive_buffer, NULL);
#endif //DMA_RESET
	AMAZON_SE_SDIO_DMSG ("amazon_se_data_recv,len=%d\n", len);
#ifndef CONFIG_TI_SDIO
	if (!data_request) {
		kfree (buf);
		return OK;
	}

	while (i < len && data_request->blocks
	       && data_request->request_size > 0) {
		memcpy (data_request->blocks->data, ((uint8_t *) buf) + i,
			1 << data_request->block_size_shift);
		i += (1 << data_request->block_size_shift);;
		data_request->request_size -=
			(1 << data_request->block_size_shift);;
		data_request->blocks = data_request->blocks->next;
	}
	kfree (buf);
	if (data_request->request_size == 0 || data_request->blocks == NULL) {
		data_access = 0;
		wake_up_interruptible
			(&wait_queue_amazon_se_sd_controller_data_cmd);
	}
#else
 memcpy (g_sdio_async_receive_buffer+g_data_recv_len, ((uint8_t *) buf),(len-g_data_len_offset));
        kfree (buf);
//      memcpy (g_sdio_async_receive_buffer+g_data_recv_len, ((uint8_t *) g_rx_receive_buffer),(len-g_data_len_offset));
        g_data_recv_len += len;
        if( g_data_recv_len  >= g_data_len )
        {
//      memcpy (g_sdio_async_receive_buffer+g_data_recv_len, ((uint8_t *) buf),(len-g_data_len_offset));
 //       kfree (buf);
//      memcpy (g_sdio_async_receive_buffer+g_data_recv_len, ((uint8_t *) g_rx_receive_buffer),(len-g_data_len_offset));
//              AMAZON_SE_SD_CONTROLLER_WRITEL (((0x6<<4)|0xe), MCI_DCTRL);
         //     AMAZON_SE_SD_CONTROLLER_WRITEL (0x0, SDIO_DMACON);
                AMAZON_SE_SD_CONTROLLER_WRITEL (AMAZON_SE_SD_CONTROLLER_READL(SDIO_DMACON) & ~SDIO_DMACON_RXON, SDIO_DMACON);
                data_read_total_size += g_data_recv_len;
                data_read_async_total_size += g_data_recv_len;
#ifndef DMA_RESET
                struct rx_desc *rx_desc_p;
                _dma_channel_info *pCh = dma_dev->rx_chan[dma_dev->current_rx_chan];
                rx_desc_p = (struct rx_desc *) pCh->desc_base + pCh->curr_desc;
                rx_desc_p->status.word = 0;
#endif
                 if (g_sdio_completion_callback != 0)
                            g_sdio_completion_callback(status);
        }
/*
        int k;
        for (  k=0; k<(len-g_data_len_offset);k++)
        {
                if (k%16 == 0) printk("\n");
                printk("%2x ",(uint8_t)g_sdio_async_receive_buffer[k]);
        }
*/
#endif  //CONFIG_TI_SDIO
	return OK;
}

static int
dma_intr_handler (struct dma_device_info *dma_dev, int status)
{

	AMAZON_SE_SDIO_DMSG ("got dma interrupt!\n");
	switch (status) {
	case RCV_INT:
		amazon_se_data_recv (dma_dev);
		break;
	case TX_BUF_FULL_INT:
		AMAZON_SE_SDIO_EMSG ("TX_BUF_FULL_INT\n");
		break;
	}
	return OK;
}

static u8 *
dma_buffer_alloc (int len, int *byte_offset, void **opt)
{
	u8 *buffer = NULL;
#ifdef CONFIG_TI_SDIO
	buffer = kmalloc (len, GFP_ATOMIC);
#else
	buffer = kmalloc (1 << amazon_se_sd_controller.blklen, GFP_ATOMIC);
#endif

	if (buffer == NULL) {
		return NULL;
	}
	//*(int*)opt=(int)buffer;
#ifndef CONFIG_TI_SDIO
	memset (buffer, 0, 1 << amazon_se_sd_controller.blklen);
#endif
	*byte_offset = 0;
	return buffer;
}

static int
dma_buffer_free (u8 * dataptr, void *opt)
{
	if (dataptr == NULL) {
	}
	else {
		//kfree(dataptr);
	}
	return OK;

}

static int
amazon_se_sdio_ioctl (struct inode *ino, struct file *fil, unsigned int command,
		   unsigned long lon)
{
	struct sd_cmd sdio_ioctl_cmd = { 0 };
	int ret = 0;
#ifdef CONFIG_INFINEON_SDIO
	sd_block_request_t sdio_ioctl_data_request;
	amazon_se_sdio_ioctl_block_request user_data_request;
	block_data_t blk_data;
#endif

	switch (command) {
	case AMAZON_SE_SDIO_SEND_CMD:
#ifdef CONFIG_INFINEON_SDIO
		copy_from_user ((char *) &sdio_ioctl_cmd, (char *) lon,
				sizeof (struct sd_cmd));
		ret = amazon_se_sd_controller_send_cmd (&amazon_se_sd_controller,
						     &sdio_ioctl_cmd);
		copy_to_user ((char *) lon, (char *) &sdio_ioctl_cmd,
			      sizeof (struct sd_cmd));
#endif
		break;
	case AMAZON_SE_SDIO_READ_DATA:
#ifdef CONFIG_INFINEON_SDIO
		copy_from_user ((char *) &user_data_request, (char *) lon,
				sizeof (amazon_se_sdio_ioctl_block_request));
		sdio_ioctl_data_request.block_size_shift =
			user_data_request.block_length;
		sdio_ioctl_data_request.nBlocks = 1;
		sdio_ioctl_data_request.request_size =
			user_data_request.data_length;
		blk_data.data = user_data_request.data;
		blk_data.next = NULL;
		sdio_ioctl_data_request.blocks = &blk_data;

		ret = amazon_se_sd_controller_read_data_pre
			(&amazon_se_sd_controller, NULL,
			 &sdio_ioctl_data_request);
		ret = amazon_se_sd_controller_send_cmd (&amazon_se_sd_controller,
						     &user_data_request.cmd);

		ret = amazon_se_sd_controller_read_data (&amazon_se_sd_controller,
						      NULL,
						      &sdio_ioctl_data_request,
						      1000);
		copy_to_user ((char *) lon, (char *) &user_data_request,
			      sizeof (amazon_se_sdio_ioctl_block_request));
#endif
		break;
	case AMAZON_SE_SDIO_SEND_DATA:
#ifdef CONFIG_INFINEON_SDIO
		copy_from_user ((char *) &user_data_request, (char *) lon,
				sizeof (amazon_se_sdio_ioctl_block_request));
		sdio_ioctl_data_request.block_size_shift =
			user_data_request.block_length;
		sdio_ioctl_data_request.nBlocks = 1;
		sdio_ioctl_data_request.request_size =
			user_data_request.data_length;
		blk_data.data = user_data_request.data;
		blk_data.next = NULL;
		sdio_ioctl_data_request.blocks = &blk_data;
		ret = amazon_se_sd_controller_send_cmd (&amazon_se_sd_controller,
						     &user_data_request.cmd);
		ret = amazon_se_sd_controller_send_data (&amazon_se_sd_controller,
						      NULL,
						      &sdio_ioctl_data_request,
						      1000);
#endif
		break;
	case AMAZON_SE_SDIO_SET_OPS_WBUS:
		if (lon == 0) {
			ret = amazon_se_sd_controller_set_ops (SD_SET_BUS_WIDTH,
							    SD_BUS_1BITS);
		}
		else if (lon == 1) {
			ret = amazon_se_sd_controller_set_ops (SD_SET_BUS_WIDTH,
							    SD_BUS_4BITS);
		}
		else {
			ret = -EINVAL;
		}
		break;
	case AMAZON_SE_SDIO_SET_OPS_FREQUENCY:
		ret = amazon_se_sd_controller_set_ops (SD_SET_FREQENCY, lon);
		break;
	case AMAZON_SE_SDIO_GET_OPS_WBUS:
		break;
	default:
		return -ENOIOCTLCMD;
	}
	return ret;
}

#if GPIO_MODE_SET 
static int
amazon_se_sdio_gpio_configure (void)
{
	uint32_t gpio_p0_dir, gpio_p1_dir;
	uint32_t gpio_p0_alt0, gpio_p1_alt0;
	uint32_t gpio_p0_alt1, gpio_p1_alt1;
	uint32_t gpio_p0_od, gpio_p1_od;
	uint32_t gpio_p0_out, gpio_p1_out;
	uint32_t gpio_p0_pudsel, gpio_p1_pudsel;
	uint32_t gpio_p0_puden, gpio_p1_puden;

	gpio_p0_out = *(AMAZON_SE_GPIO_P0_IN);
	gpio_p1_out = *(AMAZON_SE_GPIO_P1_IN);
	gpio_p0_dir = *(AMAZON_SE_GPIO_P0_DIR);
	gpio_p1_dir = *(AMAZON_SE_GPIO_P1_DIR);

	gpio_p0_alt0 = *(AMAZON_SE_GPIO_P0_ALTSEL0);
	gpio_p1_alt0 = *(AMAZON_SE_GPIO_P1_ALTSEL0);
	gpio_p0_alt1 = *(AMAZON_SE_GPIO_P0_ALTSEL1);
	gpio_p1_alt1 = *(AMAZON_SE_GPIO_P1_ALTSEL1);
	gpio_p0_od = *(AMAZON_SE_GPIO_P0_OD);
	gpio_p1_od = *(AMAZON_SE_GPIO_P1_OD);
	gpio_p0_pudsel = *(AMAZON_SE_GPIO_P0_PUDSEL);
	gpio_p1_pudsel = *(AMAZON_SE_GPIO_P1_PUDSEL);
	gpio_p0_puden = *(AMAZON_SE_GPIO_P0_PUDEN);
	gpio_p1_puden = *(AMAZON_SE_GPIO_P1_PUDEN);

	//////////////////////// MCLCLK//////////////////////////////
#if MCLCLK == 15
	//GPIO15 (P0.15) : DIR=1, ALT0=1, ALT1=1, OUT=1, OD=1
	AMAZON_SE_SDIO_DMSG ("using GPIO 15 as MCLCLK.\n");
	gpio_p0_dir |= (1 << 15 );
	gpio_p0_alt0 |= (1 <<15);
	gpio_p0_alt1 |= (1<<15);
	gpio_p0_od |= (1<<15);
	gpio_p0_pudsel |= (1<<15);
	gpio_p0_puden |= (1<<15);
#else
#error MCLCLK not defined
#endif

	/////////////////////////MCLCMD////////////////////////////
#if MCLCMD == 19
	//GPIO20 (P1.4): DIR=0,ALT0=0,ALT1=1,OUT=1, OD=1
	//GPIO19 (P1.3): DIR=0,ALT0=1,ALT1=1,OUT=1, OD=1
	AMAZON_SE_SDIO_DMSG ("using GPIO 20 as MCLCMD.\n");
	gpio_p1_dir &= ~(1 << 3);
	gpio_p1_alt0 |= (1 << 3);
	gpio_p1_alt1 |= (1 << 3);
	gpio_p1_od |= (1 << 3);
	gpio_p1_pudsel |= (1 << 3);
	gpio_p1_puden |= (1 << 3);

#else
#error MCLCMD not defined
#endif

	/////////////////////////DATA0/////////////////////////////
#if MCLDATA0 == 12
	//GPIO12 (P0.12): DIR=0,ALT0=1,ALT1=1,OUT=1, OD=1
	AMAZON_SE_SDIO_DMSG ("using GPIO 12 as DATA0.\n");
	gpio_p0_dir &= ~(1 << 12);
	gpio_p0_alt0 |= (1 << 12);
	gpio_p0_alt1 |= (1 << 12);
	gpio_p0_od |= (1 << 12);
	gpio_p0_pudsel |= (1 << 12);
	gpio_p0_puden |= (1 << 12);
#else
#error MCLDATA0 not defined
#endif
#if MCLDATA1 == 26
	//GPIO26 (P1.10): DIR=0,ALT0=1,ALT1=1,OUT=1, OD=1
	AMAZON_SE_SDIO_DMSG ("using GPIO 26 as DATA1.\n");
	gpio_p1_dir &= ~(1 << 10);
	gpio_p1_alt0 |= (1 << 10);
	gpio_p1_alt1 |= (1 << 10);
	gpio_p1_od |= (1 << 10);
	gpio_p1_pudsel |= (1 << 10);
	gpio_p1_puden |= (1 << 10);
#else
#error MCLDATA1 not defined
#endif

	/////////////////////////DATA2/////////////////////////////
#if MCLDATA2 == 28
	//GPIO26 (P1.12): DIR=0,ALT0=1,ALT1=1,OUT=1, OD=1
	AMAZON_SE_SDIO_DMSG ("using GPIO 26 as DATA2.\n");
	gpio_p1_dir &= ~(1 << 12);
	gpio_p1_alt0 |= (1 << 12);
	gpio_p1_alt1 |= (1 << 12);
	gpio_p1_od |= (1 << 12);
	gpio_p1_pudsel |= (1 << 12);
	gpio_p1_puden |= (1 << 12);
#else
#error MCLDATA2 not defined
#endif
#if MCLDATA3 == 20
	//GPIO20 (P1.4): DIR=0,ALT0=1,ALT1=1,OUT=1, OD=1
	AMAZON_SE_SDIO_DMSG ("using GPIO 25 as DATA3.\n");
	gpio_p1_dir &= ~(1 << 4);
	gpio_p1_alt0 |= (1 << 4);
	gpio_p1_alt1 |= (1 << 4);
	gpio_p1_od |= (1 << 4);
	gpio_p1_pudsel |= (1 << 4);
	gpio_p1_puden |= (1 << 4);

#else
#error MCLDATA3 not defined
#endif
	*(AMAZON_SE_GPIO_P0_DIR) = gpio_p0_dir;
	*(AMAZON_SE_GPIO_P1_DIR) = gpio_p1_dir;
	*(AMAZON_SE_GPIO_P0_ALTSEL0) = gpio_p0_alt0;
	*(AMAZON_SE_GPIO_P1_ALTSEL0) = gpio_p1_alt0;
	*(AMAZON_SE_GPIO_P0_ALTSEL1) = gpio_p0_alt1;
	*(AMAZON_SE_GPIO_P1_ALTSEL1) = gpio_p1_alt1;
	*(AMAZON_SE_GPIO_P0_OD) = gpio_p0_od;
	*(AMAZON_SE_GPIO_P1_OD) = gpio_p1_od;
	*(AMAZON_SE_GPIO_P0_OUT) = gpio_p0_out;
	*(AMAZON_SE_GPIO_P1_OUT) = gpio_p1_out;
	*(AMAZON_SE_GPIO_P0_PUDSEL) = gpio_p0_pudsel;
	*(AMAZON_SE_GPIO_P1_PUDSEL) = gpio_p1_pudsel;
	*(AMAZON_SE_GPIO_P0_PUDEN) = gpio_p0_puden;
	*(AMAZON_SE_GPIO_P1_PUDEN) = gpio_p1_puden;
	return 0;
}

#endif  //GPIO_MODE_SET
#ifdef CONFIG_TI_SDIO
/* GPIO4 is connected to the WLAN chip */
#if 1
int amazon_se_sdio_set_ext_reset(int active)
{
        uint32_t gpio_p0_dir;
        uint32_t gpio_p0_alt0;
        uint32_t gpio_p0_alt1;
        uint32_t gpio_p0_od;
        uint32_t gpio_p0_out;
        uint32_t gpio_p0_pudsel;
        uint32_t gpio_p0_puden;
	printk("%s: Line:%d",__FUNCTION__,__LINE__); 

        gpio_p0_out = *(AMAZON_SE_GPIO_P0_OUT);
        gpio_p0_dir = *(AMAZON_SE_GPIO_P0_DIR);
        gpio_p0_alt0 = *(AMAZON_SE_GPIO_P0_ALTSEL0);
        gpio_p0_alt1 = *(AMAZON_SE_GPIO_P0_ALTSEL1);
        gpio_p0_od = *(AMAZON_SE_GPIO_P0_OD);
        gpio_p0_pudsel = *(AMAZON_SE_GPIO_P0_PUDSEL);
        gpio_p0_puden = *(AMAZON_SE_GPIO_P0_PUDEN);

//        gpio_p0_out     = (gpio_p0_out & ~0x2000) | ((active == 0) ? 0x2000 : 0);
        gpio_p0_out     = (gpio_p0_out & ~0x0010) | ((active == 0) ? 0x0010 : 0);
        gpio_p0_dir    |= 0x0010;  /* output */
        gpio_p0_alt0   &= ~0x0010;
        gpio_p0_alt1   &= ~0x0010;
        gpio_p0_od     |= 0x0010;
        gpio_p0_pudsel |= 0x0010;
        gpio_p0_puden  |= 0x0010;

        *(AMAZON_SE_GPIO_P0_DIR) = gpio_p0_dir;
        *(AMAZON_SE_GPIO_P0_ALTSEL0) = gpio_p0_alt0;
        *(AMAZON_SE_GPIO_P0_ALTSEL1) = gpio_p0_alt1;
        *(AMAZON_SE_GPIO_P0_OD) = gpio_p0_od;
        *(AMAZON_SE_GPIO_P0_OUT) = gpio_p0_out;
        *(AMAZON_SE_GPIO_P0_PUDSEL) = gpio_p0_pudsel;
        *(AMAZON_SE_GPIO_P0_PUDEN) = gpio_p0_puden;
        return 0;
}
#else
/* GPIO27 is connected to the WLAN chip */
int amazon_se_sdio_set_ext_reset(int active)
{
        uint32_t gpio_p1_dir;
        uint32_t gpio_p1_alt0;
        uint32_t gpio_p1_alt1;
        uint32_t gpio_p1_od;
        uint32_t gpio_p1_out;
        uint32_t gpio_p1_pudsel;
        uint32_t gpio_p1_puden;

        gpio_p1_out = *(AMAZON_SE_GPIO_P1_OUT);
        gpio_p1_dir = *(AMAZON_SE_GPIO_P1_DIR);
        gpio_p1_alt0 = *(AMAZON_SE_GPIO_P1_ALTSEL0);
        gpio_p1_alt1 = *(AMAZON_SE_GPIO_P1_ALTSEL1);
        gpio_p1_od = *(AMAZON_SE_GPIO_P1_OD);
        gpio_p1_pudsel = *(AMAZON_SE_GPIO_P1_PUDSEL);
        gpio_p1_puden = *(AMAZON_SE_GPIO_P1_PUDEN);

        gpio_p1_out     = (gpio_p1_out & ~0x0800) | ((active == 0) ? 0 : 0x0800);
        gpio_p1_dir    |= 0x0800;  /* output */
        gpio_p1_alt0   &= ~0x0800;
        gpio_p1_alt1   &= ~0x0800;
        gpio_p1_od     |= 0x0800;
        gpio_p1_pudsel |= 0x0800;
        gpio_p1_puden  |= 0x0800;

        *(AMAZON_SE_GPIO_P1_DIR) = gpio_p1_dir;
        *(AMAZON_SE_GPIO_P1_ALTSEL0) = gpio_p1_alt0;
        *(AMAZON_SE_GPIO_P1_ALTSEL1) = gpio_p1_alt1;
        *(AMAZON_SE_GPIO_P1_OD) = gpio_p1_od;
        *(AMAZON_SE_GPIO_P1_OUT) = gpio_p1_out;
        *(AMAZON_SE_GPIO_P1_PUDSEL) = gpio_p1_pudsel;
        *(AMAZON_SE_GPIO_P1_PUDEN) = gpio_p1_puden;
    return 0;
}
#endif
#endif //CONFIG_TI_SDIO

static int
amazon_se_sdio_proc_read (struct file *file, char *buf, size_t nbytes,
		       loff_t * ppos)
{
	int i_ino = (file->f_dentry->d_inode)->i_ino;
	char outputbuf[128];
	int count = 0;
	int i;
	reg_entry_t *current_reg = NULL;
	for (i = 0; i < NUM_OF_REG_ENTRY; i++) {
		if (regs[i].low_ino == i_ino) {
			current_reg = &regs[i];
			break;
		}
	}
	if (current_reg == NULL)
		return -EINVAL;
	if (current_reg->flag == (int *) 1) {
		///proc/amazon_se_sdio/gpio
		if (*ppos > 0)	/* Assume reading completed in previous read */
			return 0;	// indicates end of file
		count = sprintf (outputbuf, "\nFunction\tGPIO Number\n");
		count += sprintf (outputbuf + count,
				  "Command:\t%d\nClock  :\t%d\nDATA 0 :\t%d\nDATA 1 :\t%d\nDATA 2 :\t%d\nDATA 3 :\t%d\n",
				  MCLCMD, MCLCLK, MCLDATA0, MCLDATA1,
				  MCLDATA2, MCLDATA3);
		*ppos += count;
	}
	else if (current_reg->flag == (int *) 2) {
		///proc/amazon_se_sdio/frequency
		if (*ppos > 0)	/* Assume reading completed in previous read */
			return 0;	// indicates end of file
		count += sprintf (outputbuf + count, "Frequency: %d\n",
				  amazon_se_sd_controller.current_speed);
		*ppos += count;
	}
	else if (current_reg->flag == (int *) 3) {
		uint32_t reg;

		if (*ppos > 0)	/* Assume reading completed in previous read */
			return 0;	// indicates end of fil

		reg = AMAZON_SE_SD_CONTROLLER_READL (MCI_CLK);
		if (reg & MCI_CLK_WD) {
			count += sprintf (outputbuf + count, "4 Bits\n");
		}
		else {
			count += sprintf (outputbuf + count, "1 Bit\n");
		}
		*ppos += count;
	}
	else {
		if (*ppos > 0)	/* Assume reading completed in previous read */
			return 0;	// indicates end of file
		count = sprintf (outputbuf, "0x%08X\n\n",
				 *(current_reg->flag));
		*ppos += count;
		if (count > nbytes)	/* Assume output can be read at one time */
			return -EINVAL;
	}

	if (copy_to_user (buf, outputbuf, count)) {
		return -EFAULT;
	}
	return count;
}

static int
setup_dma_driver (amazon_se_sd_controller_priv_t * priv)
{
	int i = 0;
	int ret;

	priv->dma_device = dma_device_reserve ("SDIO");

	if (!priv->dma_device)
		return -ENOMEM;
	priv->dma_device->intr_handler = dma_intr_handler;
	priv->dma_device->buffer_alloc = dma_buffer_alloc;
	priv->dma_device->buffer_free = dma_buffer_free;
	priv->dma_device->num_rx_chan = 1;	/*turn on all the receive channels */

	for (i = 0; i < priv->dma_device->num_rx_chan; i++) {
		priv->dma_device->rx_chan[i]->packet_size =
			1 << amazon_se_sd_controller.blklen;
		priv->dma_device->rx_chan[i]->control = AMAZON_SE_DMA_CH_ON;
	}
	ret = dma_device_register (priv->dma_device);
	for (i = 0; i < priv->dma_device->num_rx_chan; i++)
		priv->dma_device->rx_chan[i]->open (priv->dma_device->
						    rx_chan[i]);
	return ret;

}
static int
cleanup_dma_driver (amazon_se_sd_controller_priv_t * priv)
{
	if (priv == NULL)
		return -EINVAL;

	if (priv->dma_device) {
		dma_device_unregister (priv->dma_device);
		kfree (priv->dma_device);
	}
	return 0;
}

static int
amazon_se_sdio_controller_init_module (void)
{
	uint32_t sdio_id = 0;
	int ret = 0, i = 0;
	amazon_se_sd_controller_priv_t *priv;
	struct proc_dir_entry *entry;
	reg_entry_t regs_temp[PROC_ITEMS] =	// Items being debugged
	{
		/*      {       flag,          name,          description } */
		{(int *) 1, "gpio", "gpio number to used", 0},
		{(int *) 2, "frequency", "sd controller frequency", 0},
		{(int *) 3, "bus_width", "sd controller data bus width", 0},
		{&cmd_counter, "cmd_counter", "command count", 0},
		{&cmd_response_counter, "cmd_response_counter",
		 "command response count", 0},
		{&cmd_error_counter, "cmd_error_counter",
		 "command error count", 0},
		{&data_read_counter, "data_read_counter",
		 "data read command counter", 0},
		{&data_read_total_size, "data_read_total_size",
		 "total data size of reading", 0},
		{&data_write_counter, "data_write_counter",
		 "data write command counter", 0},
		{&data_write_total_size, "data_write_total_size",
		 "total data size of writing", 0},
		{&data_error_counter, "data_error_counter",
		 "data error count", 0},
		{&amazon_se_sd_controller.blklen, "block_len", "block length",
		 0},
/*
{&data_read_async_counter, "data_read_async_counter",
                 "data read command counter", 0},
                {&data_read_total_size, "data_read_total_size",
                 "total data size of reading", 0},
                {&data_read_async_total_size, "data_read_async_total_size",
                 "total data size of reading", 0},
                {&data_write_counter, "data_write_counter",
                 "data write command counter", 0},
                {&data_write_async_counter, "data_write_async_counter",
                 "data write command counter", 0},
                {&data_write_total_size, "data_write_total_size",
                 "total data size of writing", 0},
                {&data_write_async_total_size, "data_write_async_total_size",
                 "total data size of writing in async mode", 0},
                {&data_error_counter, "data_error_counter",
                 "data error count", 0},
                {&amazon_se_sd_controller.blklen, "block_len", "block length",
                 0},
                {&cmd_crc_error_counter, "cmd_crc_error_counter",
                 "command crc error counter (rx+tx+cmd)", 0},
                {&cmd_crc_rx_error_counter, "cmd_crc_rx_error_counter",
                 "command crc rx error counter", 0},
                {&cmd_crc_tx_error_counter, "cmd_crc_tx_error_counter",
                 "command crc tx error counter", 0},
                {&cmd_tx_timeout_counter, "cmd_tx_timeout_counter",
                 "Tx data command timeout counter", 0},
                {&cmd_rx_timeout_counter, "cmd_rx_timeout_counter",
                 "Rx data command timeout counter", 0},
                {&data_tx_error_counter, "data_tx_error_counter",
                 "data error count", 0},
                {&data_rx_error_counter, "data_rx_error_counter",
                 "data error count", 0},


*/
	};

	printk ("Amazon SE_sdio_controller Version:%s\n", AMAZON_SE_SDIO_VERSION);

	sdio_id = AMAZON_SE_SD_CONTROLLER_READL (SDIO_ID);
	printk ("\nAmazon SE_sdio_controller ID:%08X\n", sdio_id);
	if (sdio_id != 0xF041C030) {
		//SDIOERRMSG ("Amazon SE SDIO Controller not found!!\n");
		printk("Amazon SE SDIO Controller not found!!\n");
		return -ENODEV;
	}
#ifdef CONFIG_INFINEON_SDIO
	init_waitqueue_head (&wait_queue_amazon_se_sd_controller_cmd);
	//init_completion(&comp_amazon_se_sd_controller_cmd);
	init_waitqueue_head (&wait_queue_amazon_se_sd_controller_data_cmd);
	//init_completion(&comp_amazon_se_sd_controller_data_cmd);
#endif //CONFIG_INFINEON_SDIO
#ifdef CONFIG_TI_SDIO

                g_rx_receive_buffer = kmalloc (sizeof(char)*1024, GFP_ATOMIC);

                if(g_rx_receive_buffer == NULL)
                {
                        printk("%s:Mem allocate failed \n",__FUNCTION__);
                        return -ENOMEM;
                }

#endif

	amazon_se_sd_controller.priv =
		kmalloc (sizeof (amazon_se_sd_controller_priv_t), GFP_ATOMIC);
	if (amazon_se_sd_controller.priv == NULL)
		return -ENOMEM;
	priv = amazon_se_sd_controller.priv;

	memset (priv, 0, sizeof (amazon_se_sd_controller_priv_t));
#if 0
	// setup dma
	priv->dma_device = dma_device_reserve ("SDIO");

	if (!priv->dma_device)
		return -ENOMEM;
	priv->dma_device->intr_handler = dma_intr_handler;
	priv->dma_device->buffer_alloc = dma_buffer_alloc;
	priv->dma_device->buffer_free = dma_buffer_free;
	priv->dma_device->num_rx_chan = 1;	/*turn on all the receive channels */

	for (i = 0; i < priv->dma_device->num_rx_chan; i++) {
		priv->dma_device->rx_chan[i]->packet_size =
			1 << amazon_se_sd_controller.blklen;
		priv->dma_device->rx_chan[i]->control = AMAZON_SE_DMA_CH_ON;
	}
	ret = dma_device_register (priv->dma_device);
	for (i = 0; i < priv->dma_device->num_rx_chan; i++)
		priv->dma_device->rx_chan[i]->open (priv->dma_device->
						    rx_chan[i]);
#endif
	setup_dma_driver (priv);

	// power on SDIO module
//	*(AMAZON_SE_PMU_PWDCR) &= ~(1 << AMAZON_SE_PMU_SDIO_SHIFT);
	*(AMAZON_SE_PMU_PWDCR) &= ~(1 << 2);

	// reset sdio 
	//*(AMAZON_SE_RCU_RST_REQ) |= (AMAZON_SE_RCU_RST_REQ_SDIO);

#if GPIO_MODE_SET 
	amazon_se_sdio_gpio_configure ();
#endif  //GPIO_MODE_SET
	AMAZON_SE_SD_CONTROLLER_WRITEL (0x400, SDIO_CLC);	// 100MHz / 4 = 25MHz       

	priv->mclk_speed = 25000000;	//25   MHz

	amazon_se_sd_controller_set_ops (SD_SET_FREQENCY, 0);
	mdelay (1);
#ifdef CONFIG_INFINEON_SDIO
	amazon_se_sd_controller_set_ops (SD_SET_FREQENCY, SD_CLK_400K);
#endif

#ifdef CONFIG_TI_SDIO
//	amazon_se_sd_controller_set_ops (SD_SET_FREQENCY, SD_CLK_400K);
//        amazon_se_sd_controller_set_ops (SD_SET_FREQENCY, 25000000);
      amazon_se_sd_controller_set_ops (SD_SET_FREQENCY, 12500000);
        amazon_se_sd_controller_set_ops (SD_SET_BUS_WIDTH, SD_BUS_4BITS);
#endif  //CONFIG_TI_SDIO

	AMAZON_SE_SD_CONTROLLER_WRITEL (MCI_PWR_ON, MCI_PWR);	// Power On 


#ifdef CONFIG_TI_SDIO
        AMAZON_SE_SD_CONTROLLER_WRITEL ( SDIO_IMC_INTR1, SDIO_ICR);
        printk("SDIO Interrupt1 is enabled\n");
        ret = request_irq (AMAZON_SE_SDIO_CONTROLLER_INTR1_IRQ,
                           amazon_se_sd_controller_irq_intr1, SA_INTERRUPT,
                           DEVICE_NAME, &amazon_se_sd_controller);
        if (ret != 0) {
                SDIOERRMSG ("request_irq fail! errno:%d\n", ret);
                return ret;
        }

        AMAZON_SE_SD_CONTROLLER_WRITEL (SDIO_IMC_INTR1, SDIO_IMC);

#endif

#ifdef CONFIG_INFINEON_SDIO
	// enable DMA transmit/receive path
	AMAZON_SE_SD_CONTROLLER_WRITEL (SDIO_DMACON_TXON | SDIO_DMACON_RXON |
				     DMA_CLASS, SDIO_DMACON);

	AMAZON_SE_SD_CONTROLLER_WRITEL (SDIO_IMC_SDIO | SDIO_IMC_INTR0,
				     SDIO_ICR);

	ret = request_irq (AMAZON_SE_SDIO_CONTROLLER_I_IRQ,
			   amazon_se_sd_controller_handle_sdio_int, SA_INTERRUPT,
			   DEVICE_NAME, &amazon_se_sd_controller);
	if (ret != 0) {
		SDIOERRMSG ("request_irq fail! errno:%d\n", ret);
		return ret;
	}
	else
	printk("SDIO Interrupt0 is enabled\n");
	ret = request_irq (AMAZON_SE_SDIO_CONTROLLER_INTR0_IRQ,
			   amazon_se_sd_controller_irq, SA_INTERRUPT,
			   DEVICE_NAME, &amazon_se_sd_controller);
	if (ret != 0) {
		SDIOERRMSG ("request_irq fail! errno:%d\n", ret);
		return ret;
	}
	else
	printk("SDIO Card Interrupt is enabled\n");
	//AMAZON_SE_SD_CONTROLLER_WRITEL(SDIO_IMC_SDIO|SDIO_IMC_INTR0,SDIO_IMC);
	AMAZON_SE_SD_CONTROLLER_WRITEL (SDIO_IMC_INTR0, SDIO_IMC);

#endif //CONFIG_INFINEON_SDIO
	// clear all interrupt flags
	AMAZON_SE_SD_CONTROLLER_WRITEL (0x7FF, MCI_CL);

	// disable MMC interrupt
	AMAZON_SE_SD_CONTROLLER_WRITEL (0, MCI_IM0);
	AMAZON_SE_SD_CONTROLLER_WRITEL (0, MCI_IM1);


	// enable SDIO
#if GPIO_MODE_SET 
	printk("\n SDIO  with GPIO Mode ");
	AMAZON_SE_SD_CONTROLLER_WRITEL (SDIO_CTRL_SDIOEN, SDIO_CTRL);
	AMAZON_SE_SDIO_DMSG ("\nSDIO_CTRL=%08X\n",
				  AMAZON_SE_SD_CONTROLLER_READL (SDIO_CTRL));

#else
	printk("\n SDIO  with Dedicated Mode ");
	AMAZON_SE_SD_CONTROLLER_WRITEL (SDIO_CTRL_SDIOEN | BIT1, SDIO_CTRL);
#endif
	AMAZON_SE_SDIO_DMSG ("\nSDIO_CTRL1   =%08X\n",
				  AMAZON_SE_SD_CONTROLLER_READL (SDIO_CTRL));
	AMAZON_SE_SDIO_DMSG ("\nSDIO_CLC=%08X\n",
				  AMAZON_SE_SD_CONTROLLER_READL (SDIO_CLC));
	AMAZON_SE_SDIO_DMSG ("\nMCI_CLK=%08X\n",
				  AMAZON_SE_SD_CONTROLLER_READL (MCI_CLK));
#ifdef CONFIG_INFINEON_SDIO
#ifdef CONFIG_SD_MEMORY
	register_sd_controller (&amazon_se_sd_controller);
#endif
#endif
#ifdef CONFIG_DEVFS_FS
	if (devfs_register_chrdev
	    (major, "amazon_se_sdio", &amazon_se_sdio_operations) != 0) {
		AMAZON_SE_SDIO_EMSG
			("\n\n unable to register major for amazon_se_sdio!!!");
		return -1;
	}
#else
	if (register_chrdev (major, "amazon_se_sdio", &amazon_se_sdio_operations) !=
	    0) {
		AMAZON_SE_SDIO_EMSG
			("\n\n unable to register major for amazon_se_sdio!!!");
		return -1;
	}
#endif

	// procfs
	amazon_se_sdio_dir = proc_mkdir (AMAZON_SE_SDIO_PROC_DIRNAME, NULL);
	if (amazon_se_sdio_dir == NULL) {
		printk (KERN_ERR ": can't create /proc/"
			AMAZON_SE_SDIO_PROC_DIRNAME "\n\n");
		return (-ENOMEM);
	}

	memcpy ((char *) regs, (char *) regs_temp, sizeof (regs_temp));

	for (i = 0; i < NUM_OF_REG_ENTRY; i++) {
		entry = create_proc_entry (regs[i].name,
					   S_IWUSR | S_IRUSR | S_IRGRP |
					   S_IROTH, amazon_se_sdio_dir);
		if (entry) {
			regs[i].low_ino = entry->low_ino;
			entry->proc_fops = &proc_operations;
		}
		else {
			printk (KERN_ERR
				": can't create /proc/"
				AMAZON_SE_SDIO_PROC_DIRNAME "/%s\n\n",
				regs[i].name);
			return (-ENOMEM);
		}
	}
	return 0;
}

static void
amazon_se_sdio_controller__exit_module (void)
{
	amazon_se_sd_controller_priv_t *priv;
#ifdef CONFIG_PROC_FS
	int i = 0;
#endif

	SDIODBGMSG ("Module amazon_se_sdio_controller exit\n");

#ifdef CONFIG_PROC_FS
	for (i = 0; i < NUM_OF_REG_ENTRY; i++)
		remove_proc_entry (regs[i].name, amazon_se_sdio_dir);

	remove_proc_entry (AMAZON_SE_SDIO_PROC_DIRNAME, &proc_root);
#endif //CONFIG_PROC_FS

#ifdef CONFIG_INFINEON_SDIO
	free_irq (AMAZON_SE_SDIO_CONTROLLER_I_IRQ, NULL);
	free_irq (AMAZON_SE_SDIO_CONTROLLER_INTR0_IRQ, NULL);
#endif
#ifdef CONFIG_TI_SDIO
	free_irq (AMAZON_SE_SDIO_CONTROLLER_INTR1_IRQ, NULL);
#endif
#ifdef CONFIG_INFINEON_SDIO
#ifdef CONFIG_SD_MEMORY
	unregister_sd_controller (&amazon_se_sd_controller);
#endif
#endif

	priv = amazon_se_sd_controller.priv;
#if 0
	if (priv->dma_device) {
		dma_device_unregister (priv->dma_device);
		kfree (priv->dma_device);
	}
#endif
	cleanup_dma_driver (priv);

	kfree (priv);
#ifdef CONFIG_DEVFS_FS
	devfs_unregister_chrdev (major, "amazon_se_sdio");
#else
	unregister_chrdev (major, "amazon_se_sdio");
#endif
#ifdef CONFIG_INFINEON_SDIO
#ifdef CONFIG_SD_MEMORY
	unregister_sd_controller (&amazon_se_sd_controller);
#endif
#endif

	priv = amazon_se_sd_controller.priv;
#if 0
	if (priv->dma_device) {
		dma_device_unregister (priv->dma_device);
		kfree (priv->dma_device);
	}
#endif
	cleanup_dma_driver (priv);

	kfree (priv);
#ifdef CONFIG_DEVFS_FS
	devfs_unregister_chrdev (major, "amazon_se_sdio");
#else
	unregister_chrdev (major, "amazon_se_sdio");
#endif
}


#ifdef CONFIG_TI_SDIO
void *amazon_se_sd_controller_register_callback(void (*fct)(unsigned long))
{

    g_sdio_completion_callback = fct;
}
#endif


module_init (amazon_se_sdio_controller_init_module);
module_exit (amazon_se_sdio_controller__exit_module);

EXPORT_SYMBOL(amazon_se_sd_controller_set_ops);
#ifdef CONFIG_TI_SDIO
EXPORT_SYMBOL(amazon_se_sd_controller_send_data_kernel);
EXPORT_SYMBOL(amazon_se_sd_controller_read_data_kernel);
EXPORT_SYMBOL(amazon_se_sd_controller_read_async_data_kernel);
EXPORT_SYMBOL(amazon_se_sd_controller_send_async_data_kernel);
EXPORT_SYMBOL(amazon_se_sd_controller_send_cmd_kernel);
EXPORT_SYMBOL(amazon_se_sdio_set_ext_reset);
EXPORT_SYMBOL(dump_data_and_time);
EXPORT_SYMBOL(amazon_se_sd_controller_register_callback);
#endif

